/*!
 * \file afilterctl.c
 * \brief Tool for controlling advanced filter
 * \author Lukas Kekely <xkekel00@stud.fit.vutbr.cz>
 * \date 2012
 *
 * Copyright (C) 2012 Brno University of Technology
 *
 * LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: afilterctl.c 1845 2012-07-18 11:51:58Z xkekel00 $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include "hwio.h"
#include "afilterctl_input.h"

#define DTREE_PATH "/proc/device-tree"
#define INTERFACES    2    // number of interfaces

/// print help message
void usage() {
    printf("\
Usage: afilterctl [-cChirsS] [-I v] [-a t;k;d] [-d t;k] [-x file]\n\
    -a t;k;d  Add specified item.\n\
    -c        Display packet counters.\n\
    -C        Display and reset packet counters.\n\
    -d t;k    Delete items with specified type and key.\n\
    -h        This help message.\n\
    -i        Show currently used initialization vector.\n\
    -I v      Set initialization vector to given value (also clear rules).\n\
    -r        Show all rules.\n\
    -R        Show and reset (clear) all rules.\n\
    -s        Show status of tables with rules.\n\
    -x file   Device tree or design.xml path.\n\
Where:\n\
    t = Rule type, letter from <a,i> or <A,I> or number <1,5>.\n\
    k = Rule key, 5-tuple \"srcip;dstip;srcport;dstport;proto\".\n\
        Some parts of 5-tuple are ignored depending on rule type.\n\
        Ignored parts of 5-tuple can be blank.\n\
    d = Rule data (id), 32-bit unsined integer.\n\
    v = Init vectors, sequence of 32-bit unsigned integers.\n\
");
}


/// initialize communication with HW filter
int filter_init(const char *fn) {
    int x;
    x=hwio_init(fn,"Advanced_filter");
    if(x) {
        fprintf(stderr,"Unable to initialize filter [hwio#%d]!\n",x);
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}


/// deinitialize communication with HW filter
void filter_deinit() {
    hwio_deinit();
}


/// read HW filter counters
int filter_counters(int reset) {
    flt_cnt_status s[INTERFACES];
    if(flt_read_counters(reset,INTERFACES,s)) {
        fprintf(stderr,"Error reading data from HW!\n");
        return EXIT_FAILURE;
    }
    printf(" *** PACKET COUNTERS *** \n");
    for(unsigned int i=0; i<INTERFACES; i++) {
        printf("Interface %d:\n",i);
        printf("    Denied : %10" PRIu64 "\n",s[i].deny);
        printf("    Allowed: %10" PRIu64 "\n",s[i].allow);
        printf("    TOTAL  : %10" PRIu64 "\n",s[i].allow+s[i].deny);
    }
    printf("\n");
    return EXIT_SUCCESS;
}


/// add rule to HW filter
int filter_add(const char *str) {
    int size=strlen(str)+1,x;
    char sstr[size];
    flt_rule r;
    memcpy(sstr,str,size);
    if((x=flt_parse_rule(sstr,&r))!=0) {
        if(x==PARSE_RULE_ERROR)
            fprintf(stderr,"Wrong rule format, syntax error!\n");
        if(x==PARSE_RULE_EMPTY)
            fprintf(stderr,"Wrong rule format, required value left empty!\n");
        return EXIT_FAILURE;
    }
    if((x=flt_add_rule(&r))!=0) {
        if(x==ADD_RULE_BUSY)
            fprintf(stderr,"Error in HW filter, possible infinite busy state!\n");
        if(x==ADD_RULE_FULL)
            fprintf(stderr,"Table with desired rule type already full!\n");
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}


/// delete rule from HW filter
int filter_del(const char *str) {
    int size=strlen(str)+1,x;
    char sstr[size+2];
    flt_rule r;
    memcpy(sstr,str,size);
    sstr[size-1]=';';
    sstr[size]='0';
    sstr[size+1]='\0';
    if((x=flt_parse_rule(sstr,&r))!=0) {
        if(x==PARSE_RULE_ERROR)
            fprintf(stderr,"Wrong rule format, syntax error!\n");
        if(x==PARSE_RULE_EMPTY)
            fprintf(stderr,"Wrong rule format, required value left empty!\n");
        return EXIT_FAILURE;
    }
    if((x=flt_del_rule(&r))!=0) {
        if(x==DEL_RULE_BUSY)
            fprintf(stderr,"Error in HW filter, possible infinite busy state!\n");
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}


/// read HW filter status
int filter_status() {
    const char *busy="(BUSY)";
    const char *ready="(READY)";
    const char *full="(FULL)";
    const char *ignore=" (IGNORED)";
    const char *empty="";
    flt_status_flags s;
    flt_read_status(&s);
    printf(" *** FILTER STATUS *** \n");
    printf("STATUS FLAGS : %08x\n",s.vector);
    printf("Filter config: %s\n",(s.busy)?busy:ready);
    printf("Group AB     : %s%s\n",(s.busy_ab)?busy:((s.full_ab)?full:ready),(s.ignore_ab)?ignore:empty);
    printf("Group CD     : %s%s\n",(s.busy_cd)?busy:((s.full_cd)?full:ready),(s.ignore_cd)?ignore:empty);
    printf("Group EF     : %s%s\n",(s.busy_ef)?busy:((s.full_ef)?full:ready),(s.ignore_ef)?ignore:empty);
    printf("Group G      : %s%s\n",(s.busy_g )?busy:((s.full_g )?full:ready),(s.ignore_g )?ignore:empty);
    printf("Group HI     : %s%s\n",(s.busy_hi)?busy:((s.full_hi)?full:ready),(s.ignore_hi)?ignore:empty);
    printf("\n");
    return EXIT_SUCCESS;
}


/// read/reset HW filter rules
int filter_rules(int read, int reset) {
    if(!read&&!reset)
        return EXIT_SUCCESS;
    unsigned ignm=(reset)?0x1F:0;
    if(!flt_wait_ready()) {  // wait for filter ready
        fprintf(stderr,"Error in HW filter, possible infinite busy state!\n");
        return EXIT_FAILURE;
    }
    flt_stop_cmd(0x1F,ignm); // stop filter reconfiguration operations and possibly soft clear tables (ignore results)
    if(!flt_wait_ready()) { // wait until stop_cmd is finished
        fprintf(stderr,"Error in HW filter, possible infinite busy state!\n");
        return EXIT_FAILURE;
    }
    flt_dir_iterator it;
    flt_rule r;
    char str[IPSTR_LEN];
    if(read)
        printf(" *** RULES INSIDE FILTER *** \n");
    for(unsigned type=RULE_TYPE_AB; type<=RULE_TYPE_HI; type++) { // each rule type
        flt_dir_iterator_init(type,&it);
        if(!read)
            flt_dir_iterator_clearall(&it); // fast clean
        else
            while(flt_dir_iterator_nextrule(&it,&r)) {
                switch(type) {
                    case RULE_TYPE_AB:
                        ip2string((unsigned char *)r.srcip,str);
                        printf("1;%s;;;;;%u\n",str,r.id);
                        break;
                    case RULE_TYPE_CD:
                        ip2string((unsigned char *)r.srcip,str);
                        printf("2;%s/64;;;;;%u\n",str,r.id);
                        break;
                    case RULE_TYPE_EF:
                        ip2string((unsigned char *)r.srcip,str);
                        printf("3;%s",str);
                        ip2string((unsigned char *)r.dstip,str);
                        printf(";%s;%u;%u;%u;%u\n",str,r.srcport,r.dstport,r.proto,r.id);
                        break;
                    case RULE_TYPE_G:
                        printf("4;;;%u;;%u;%u\n",r.srcport,r.proto,r.id);
                        break;
                    case RULE_TYPE_HI:
                        ip2string((unsigned char *)r.srcip,str);
                        printf("5;%s;;%u;;%u;%u\n",str,r.srcport,r.proto,r.id);
                        break;
                }
                if(reset)
                    flt_dir_iterator_clear(&it); // clear rule
                flt_dir_iterator_next(&it);
            }
    }
    if(read)
        printf("\n");
    flt_stop_cmd(0,0); // enable all
    return EXIT_SUCCESS;
}

/// read HW filter init vectors
int filter_read_iv() {
    uint32_t iv[FILTER_IV_WORDS];
    flt_read_initv(iv);
    printf(" *** INITIALIZATION VECTOR *** \n");
    for(int i=0; i<FILTER_IV_WORDS; i++)
        printf("Seed %d: %08x\n",i,iv[i]);
    printf("\n");
    return EXIT_SUCCESS;
}


/// set HW filter init vectors
int filter_write_iv(char *str) {
    uint32_t iv[FILTER_IV_WORDS];
    int x;
    if((x=flt_parse_initv(str,iv))!=0) {
        if(x==PARSE_INITV_ERROR)
            fprintf(stderr,"Wrong init vector format, expected format '%%u;%%u;...'!\n");
        if(x==PARSE_INITV_NOEFF)
            fprintf(stderr,"Wrong init vector format, no effect!\n");
        return EXIT_FAILURE;
    }
    if(filter_rules(0,1)!=EXIT_SUCCESS) // silently reset rules
        return EXIT_FAILURE;
    flt_write_initv(iv);
    return EXIT_SUCCESS;
}



/// MAIN //////////////////////////////////////////////////////
int main(int argc, char *argv[]) {
    argc--;
    argv++;
    int ret=EXIT_SUCCESS, need_arg=0;
    int flag_c=0, flag_bc=0, flag_h=0, flag_r=0, flag_s=0, flag_br=0, flag_i=0;
    char *a=NULL, *d=NULL, *x=NULL, *I=NULL, **arg=NULL;
    if(!argc)
        flag_h=1;

    /// get command line options
    while(argc) {
        need_arg=0;
        if((*argv)[0]!='-'||strlen((*argv))!=2) {
            fprintf(stderr,"Unexpected parameter '%s'!\n",*argv);
            exit(1);
        }
        switch((*argv)[1]) {
            case 'a':   need_arg=1; arg=&a;   // add rule command
                        break;
            case 'c':   flag_c=1; flag_bc=0;  // read counters only
                        break;
            case 'C':   flag_c=0; flag_bc=1;  // read counters and reset
                        break;
            case 'd':   need_arg=1; arg=&d;   // delete rule command
                        break;
            case 'h':   flag_h=1;             // help msg
                        break;
            case 'i':   flag_i=1; I=NULL;     // read seeds
                        break;
            case 'I':   flag_i=0; arg=&I;     // set seeds
                        need_arg=1;
                        break;
            case 'r':   flag_r=1; flag_br=0;  // read rules
                        break;
            case 'R':   flag_r=0; flag_br=1;  // read and reset rules
                        break;
            case 's':   flag_s=1;             // show status
                        break;
            case 'x':   need_arg=1; arg=&x;   // path to Design.xml
                        break;
            default:    fprintf(stderr,"Unknown option '%s'!\n",*argv);
                        exit(1);
        }
        if(need_arg) {
            if(argc>1) {
                if(*arg!=NULL) {
                    fprintf(stderr,"Multiple definitions of parameter '%s' with argument!\n",*argv);
                    exit(1);
                }
                *arg=argv[1];
                argv++;
                argc--;
            } else {
                fprintf(stderr,"Option '%s' requires an argument!\n",*argv);
                exit(1);
            }
        }
        argc--;
        argv++;
    }

    /// check command line options
    if((a&&(d||flag_br||I))||(d&&(flag_br||I))||(flag_br&&I)) {
        fprintf(stderr,"Only one from '-a', '-d', '-R', '-I' can be defined!\n");
        exit(1);
    }

    if(!x)
        x=DTREE_PATH;

    /// execute desired operations
    ret=EXIT_FAILURE;
    if(flag_h) {
        usage();
        ret=EXIT_SUCCESS;
    } else if(filter_init(x)==EXIT_SUCCESS) {
        ret=EXIT_SUCCESS;
        if(a&&ret==EXIT_SUCCESS)
            ret=filter_add(a);
        if(d&&ret==EXIT_SUCCESS)
            ret=filter_del(d);
        if(flag_i&&ret==EXIT_SUCCESS)
            ret=filter_read_iv();
        if(I&&ret==EXIT_SUCCESS)
            ret=filter_write_iv(I);
        if((flag_r||flag_br)&&ret==EXIT_SUCCESS)
            ret=filter_rules(1,flag_br);
        if((flag_c||flag_bc)&&ret==EXIT_SUCCESS)
            ret=filter_counters(flag_bc);
        if(flag_s&&ret==EXIT_SUCCESS)
            ret=filter_status();
        filter_deinit();
    }
    return ret;
}
