/*!
 * \file afilterctl_main.c
 * \brief Tool for controlling advanced filter - configure filter
 * \author Lukas Kekely <xkekel00@stud.fit.vutbr.cz>
 * \date 2012
 *
 * Copyright (C) 2012 Brno University of Technology
 *
 * LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: afilterctl_main.c 1745 2012-07-10 22:06:36Z xkekel00 $
 *
 */

#include "afilterctl_main.h"
#include "afilterctl_input.h"
#include "hwio.h"
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/// offset of status registers
#define STATUS_OFFSET  0x200

/// filter registers addresses
#define REG_STATUS     0x000
#define REG_STOPCMD    0x004
#define REG_ADDCMD     0x008
#define REG_DELCMD     0x00C
#define REG_DIRRD      0x010
#define REG_DIRCLR     0x014
#define REG_DATA       0x018
#define REG_PROTO      0x01C
#define REG_PORTS      0x020
#define REG_SRCIP      0x024
#define REG_DSTIP      (REG_SRCIP+(FILTER_IP_WORDS<<2))
#define REG_IV         (REG_DSTIP+(FILTER_IP_WORDS<<2))

/// status flags masks
#define MASK_BUSY      0x00000001
#define MASK_BUSY_AB   0x00000020
#define MASK_BUSY_CD   0x00000010
#define MASK_BUSY_EF   0x00000008
#define MASK_BUSY_G    0x00000004
#define MASK_BUSY_HI   0x00000002
#define MASK_FULL_AB   0x00000400
#define MASK_FULL_CD   0x00000200
#define MASK_FULL_EF   0x00000100
#define MASK_FULL_G    0x00000080
#define MASK_FULL_HI   0x00000040
#define MASK_IGNORE_AB   0x00008000
#define MASK_IGNORE_CD   0x00004000
#define MASK_IGNORE_EF   0x00002000
#define MASK_IGNORE_G    0x00001000
#define MASK_IGNORE_HI   0x00000800

/// read counters of given interface
inline void flt_read_ifc_counters(int iface, flt_cnt_status *s) {
    uint64_t pom;
    pom=hwio_read32(STATUS_OFFSET+(iface<<4)+0x4);
    pom=pom<<32;
    pom+=hwio_read32(STATUS_OFFSET+(iface<<4));
    s->allow=pom;
    pom=hwio_read32(STATUS_OFFSET+(iface<<4)+0xC);
    pom=pom<<32;
    pom+=hwio_read32(STATUS_OFFSET+(iface<<4)+0x8);
    s->deny=pom;
}

/// reset counters of given interface
inline void flt_reset_ifc_counters(int iface) {
    hwio_write32(STATUS_OFFSET+(iface<<4),0);
}

/// read counters of each interface of HW filter
int flt_read_counters(int reset, int ints, flt_cnt_status *res) {
    if(ints<=0)
        return 0;
    for(int i=0; i<ints; i++) {
        flt_read_ifc_counters(i,res+i);
        if(reset)
            flt_reset_ifc_counters(i);
    }
    return 0;
}

/// read status of HW filter rules
int flt_read_status(flt_status_flags *res) {
    uint32_t pom=hwio_read32(REG_STATUS);
    res->vector=pom;
    res->busy=(pom&MASK_BUSY)&&1;
    res->busy_ab=(pom&MASK_BUSY_AB)&&1;
    res->busy_cd=(pom&MASK_BUSY_CD)&&1;
    res->busy_ef=(pom&MASK_BUSY_EF)&&1;
    res->busy_g =(pom&MASK_BUSY_G)&&1;
    res->busy_hi=(pom&MASK_BUSY_HI)&&1;
    res->full_ab=(pom&MASK_FULL_AB)&&1;
    res->full_cd=(pom&MASK_FULL_CD)&&1;
    res->full_ef=(pom&MASK_FULL_EF)&&1;
    res->full_g =(pom&MASK_FULL_G)&&1;
    res->full_hi=(pom&MASK_FULL_HI)&&1;
    res->ignore_ab=(pom&MASK_IGNORE_AB)&&1;
    res->ignore_cd=(pom&MASK_IGNORE_CD)&&1;
    res->ignore_ef=(pom&MASK_IGNORE_EF)&&1;
    res->ignore_g=(pom&MASK_IGNORE_G)&&1;
    res->ignore_hi=(pom&MASK_IGNORE_HI)&&1;
    return 0;
}

/// parse rule from string into rule structure
int flt_parse_rule(char *str, flt_rule *res) {
    unsigned e;
    if(!flt_get_rule(str,res,&e))
        return PARSE_RULE_ERROR;
    switch(res->type) {
        case RULE_TYPE_AB:
            if((e&(EMPTY_SRCIP|EMPTY_DATA))!=0)
                return PARSE_RULE_EMPTY;
            break;
        case RULE_TYPE_CD:
            if((e&(EMPTY_SRCIP|EMPTY_DATA))!=0)
                return PARSE_RULE_EMPTY;
            break;
        case RULE_TYPE_EF:
            if((e&(EMPTY_SRCIP|EMPTY_DSTIP|EMPTY_SRCPORT|EMPTY_DSTPORT|EMPTY_PROTO|EMPTY_DATA))!=0)
                return PARSE_RULE_EMPTY;
            break;
        case RULE_TYPE_G:
            if((e&(EMPTY_SRCPORT|EMPTY_PROTO|EMPTY_DATA))!=0)
                return PARSE_RULE_EMPTY;
            break;
        case RULE_TYPE_HI:
            if((e&(EMPTY_SRCIP|EMPTY_SRCPORT|EMPTY_PROTO|EMPTY_DATA))!=0)
                return PARSE_RULE_EMPTY;
            break;
        default:
            return PARSE_RULE_ERROR;
    }
    return 0;
}

/// fills register with source IP address
int flt_write_srcip(uint32_t *ip) {
    uint32_t addr=REG_SRCIP;
    for(int i=0; i<FILTER_IP_WORDS; i++, addr+=4) {
        hwio_write32(addr,ip[i]);
    }
    return 0;
}
/// fills register with destination IP address
int flt_write_dstip(uint32_t *ip) {
    uint32_t addr=REG_DSTIP;
    for(int i=0; i<FILTER_IP_WORDS; i++, addr+=4) {
        hwio_write32(addr,ip[i]);
    }
    return 0;
}
/// fills register with port numbers
int flt_write_ports(uint16_t sp, uint16_t dp) {
    hwio_write32(REG_PORTS,(((uint32_t)dp)<<16)+sp);
    return 0;
}
/// fills register with protocol number
int flt_write_proto(uint8_t p) {
    hwio_write32(REG_PROTO,p);
    return 0;
}
/// fills register with rule identificator
int flt_write_data(uint32_t id) {
    hwio_write32(REG_DATA,id);
    return 0;
}

/// reads register with source IP address
int flt_read_srcip(uint32_t *ip) {
    uint32_t addr=REG_SRCIP;
    for(int i=0; i<FILTER_IP_WORDS; i++, addr+=4) {
        ip[i]=hwio_read32(addr);
    }
    return 0;
}
/// reads register with destination IP address
int flt_read_dstip(uint32_t *ip) {
    uint32_t addr=REG_DSTIP;
    for(int i=0; i<FILTER_IP_WORDS; i++, addr+=4) {
        ip[i]=hwio_read32(addr);
    }
    return 0;
}
/// reads register with port numbers
int flt_read_ports(uint16_t *sp, uint16_t *dp) {
    uint32_t pom=hwio_read32(REG_PORTS);
    if(sp) *sp=pom&0xFFFF;
    if(dp) *dp=(pom>>16)&0xFFFF;
    return 0;
}
/// reads register with protocol number
int flt_read_proto(uint8_t *p) {
    *p=hwio_read32(REG_PROTO)&0xFF;
    return 0;
}
/// reads register with rule identificator
int flt_read_data(uint32_t *id) {
    *id=hwio_read32(REG_DATA);
    return 0;
}

/// fills data registers of HW filter
int flt_write_registers(flt_rule *r) {
    switch(r->type) {
        case RULE_TYPE_AB:
            flt_write_srcip(r->srcip);
            flt_write_data(r->id);
            break;
        case RULE_TYPE_CD:
            flt_write_srcip(r->srcip);
            flt_write_data(r->id);
            break;
        case RULE_TYPE_EF:
            flt_write_srcip(r->srcip);
            flt_write_dstip(r->dstip);
            flt_write_ports(r->srcport,r->dstport);
            flt_write_proto(r->proto);
            flt_write_data(r->id);
            break;
        case RULE_TYPE_G:
            flt_write_ports(r->srcport,0);
            flt_write_proto(r->proto);
            flt_write_data(r->id);
            break;
        case RULE_TYPE_HI:
            flt_write_srcip(r->srcip);
            flt_write_ports(r->srcport,0);
            flt_write_proto(r->proto);
            flt_write_data(r->id);
            break;
    }
    return 0;
}

/// reads data registers of HW filter
int flt_read_registers(flt_rule *r) {
    switch(r->type) {
        case RULE_TYPE_AB:
            flt_read_srcip(r->srcip);
            flt_read_data(&(r->id));
            break;
        case RULE_TYPE_CD:
            flt_read_srcip(r->srcip);
            flt_read_data(&(r->id));
            break;
        case RULE_TYPE_EF:
            flt_read_srcip(r->srcip);
            flt_read_dstip(r->dstip);
            flt_read_ports(&(r->srcport),&(r->dstport));
            flt_read_proto(&(r->proto));
            flt_read_data(&(r->id));
            break;
        case RULE_TYPE_G:
            flt_read_ports(&(r->srcport),NULL);
            flt_read_proto(&(r->proto));
            flt_read_data(&(r->id));
            break;
        case RULE_TYPE_HI:
            flt_read_srcip(r->srcip);
            flt_read_ports(&(r->srcport),NULL);
            flt_read_proto(&(r->proto));
            flt_read_data(&(r->id));
            break;
    }
    return 0;
}

/// calls add command of HW filter
int flt_add_rule_cmd(unsigned mask) {
    hwio_write32(REG_ADDCMD,mask);
    return 0;
}

/// calls remove command of HW filter
int flt_del_rule_cmd(unsigned mask) {
    hwio_write32(REG_DELCMD,mask);
    return 0;
}

/// wait for filter to be ready
int flt_wait_ready() {
    int loops=0;
    while(hwio_read32(REG_STATUS)&MASK_BUSY) {
        loops++;
        if(loops>=1024)
            return 0;
        usleep(100);
    }
    return 1;
}

/// adds rule into HW filter
int flt_add_rule(flt_rule *r) {
    uint32_t stat_mask=0, add_mask=0;
    int loops=0;
    // wait for filter to be ready
    if(!flt_wait_ready())
        return ADD_RULE_BUSY;
    // fill registers
    flt_write_registers(r);
    // masks
    switch(r->type) {
        case RULE_TYPE_AB:
            stat_mask=MASK_FULL_AB;
            add_mask=MASK_AB;
            break;
        case RULE_TYPE_CD:
            stat_mask=MASK_FULL_CD;
            add_mask=MASK_CD;
            break;
        case RULE_TYPE_EF:
            stat_mask=MASK_FULL_EF;
            add_mask=MASK_EF;
            break;
        case RULE_TYPE_G:
            stat_mask=MASK_FULL_G;
            add_mask=MASK_G;
            break;
        case RULE_TYPE_HI:
            stat_mask=MASK_FULL_HI;
            add_mask=MASK_HI;
            break;
    }
    // wait for end of last add
    while(hwio_read32(REG_STATUS)&stat_mask) {
        loops++;
        if(loops>=1024)
            return ADD_RULE_FULL;
        usleep(100);
    }
    // add rule
    flt_add_rule_cmd(add_mask);
    return 0;
}

/// removes rule from HW filter
int flt_del_rule(flt_rule *r) {
    uint32_t del_mask=0;
    // wait for filter to be ready
    if(!flt_wait_ready())
        return DEL_RULE_BUSY;
    // fill registers
    flt_write_registers(r);
    // masks
    switch(r->type) {
        case RULE_TYPE_AB:
            del_mask=MASK_AB;
            break;
        case RULE_TYPE_CD:
            del_mask=MASK_CD;
            break;
        case RULE_TYPE_EF:
            del_mask=MASK_EF;
            break;
        case RULE_TYPE_G:
            del_mask=MASK_G;
            break;
        case RULE_TYPE_HI:
            del_mask=MASK_HI;
            break;
    }
    // remove rule
    flt_del_rule_cmd(del_mask);
    return 0;
}

/// reads init vector register from HW filter
int flt_read_initv(uint32_t *iv) {
    uint32_t addr=REG_IV;
    for(int i=0; i<FILTER_IV_WORDS; i++, addr+=4)
        iv[i]=hwio_read32(addr);
    return 0;
}

/// fills init vector register inside HW filter
int flt_write_initv(uint32_t *iv) {
    uint32_t addr=REG_IV;
    for(int i=0; i<FILTER_IV_WORDS; i++, addr+=4)
        hwio_write32(addr,iv[i]);
    return 0;
}

/// parse init vector register from string representation
int flt_parse_initv(char *str, uint32_t *iv) {
    unsigned e;
    if(!flt_get_initv(str,iv,&e))
        return PARSE_INITV_ERROR;
    if(e==((1<<FILTER_IV_WORDS)-1))
        return PARSE_INITV_NOEFF;
    for(int i=0; i<FILTER_IV_WORDS; i++)
        if(e&(1<<i))
            iv[i]=hwio_read32(REG_IV+4*i);
    return 0;
}

/// calls stop command of HW filter
int flt_stop_cmd(unsigned mask_stop_add, unsigned mask_ignore_res) {
    hwio_write32(REG_STOPCMD,(mask_stop_add<<5)|mask_ignore_res);
    return 0;
}

/// initialize direct iterator into HW filter hash tables
void flt_dir_iterator_init(int rt, flt_dir_iterator *it) {
    it->type=rt-1;
    it->table=0;
    it->item=0;
}

/// next iterator, should be empty field
void flt_dir_iterator_next(flt_dir_iterator *it) {
    it->item++;
    if(FILTER_RULES_ITEMS[it->type]<=it->item) {
        it->item=0;
        it->table++;
    }
}

/// transforms dirrect iterator into dirrect address
uint32_t flt_dir_iterator_2addr(flt_dir_iterator *it) {
    return (it->type<<29)|(it->table<<FILTER_RULES_ITEMS_WIDTH[it->type])|(it->item);
}

/// finds and reads next valid rule from HW filter hash tables
int flt_dir_iterator_nextrule(flt_dir_iterator *it, flt_rule *r) {
    uint32_t pom;
    uint32_t tables=FILTER_RULES_TABLES[it->type];
    while(it->table<tables||(it->table==tables&&it->item==0)) {
        hwio_write32(REG_DIRRD,flt_dir_iterator_2addr(it));
        flt_read_data(&pom);
        if(pom!=FILTER_NOMATCH) {
            r->type=it->type+1;
            flt_read_registers(r);
            return 1;
        }
        flt_dir_iterator_next(it);
    }
    return 0;
}

/// clears field inside HW filter hash tables at address given by iterator
void flt_dir_iterator_clear(flt_dir_iterator *it) {
    hwio_write32(REG_DIRCLR,flt_dir_iterator_2addr(it));
}

/// clears all rules inside HW filter hash tables of type given by iterator
void flt_dir_iterator_clearall(flt_dir_iterator *it) {
    uint32_t items=FILTER_RULES_ITEMS[it->type];
    int l2items=FILTER_RULES_ITEMS_WIDTH[it->type];
    uint32_t tables=FILTER_RULES_TABLES[it->type];
    uint32_t start=it->type<<29;
    uint32_t table_start;
    for(uint32_t t=0; t<tables; t++) {
        table_start=start|(t<<l2items);
        for(uint32_t i=0; i<items; i++)
            hwio_write32(REG_DIRCLR,table_start|i);
    }
    hwio_write32(REG_DIRCLR,start|(tables<<l2items));
}

