#! /usr/bin/env python

import sys
import socket
import struct
import binascii
import time
import yaml

def server_create(port, host = None, backlog = 5):
	s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

	if host is None:
		s.bind((socket.gethostname(), port))
	else:
	 	s.bind((host, port))

	s.listen(backlog)
	return s

def reverse(v):
	return v[::-1]

def parse_ipv4(addr):
	octets = addr.split(".")
	if len(octets) != 4:
		raise RuntimeError("IPv4 address must be in format a.b.c.d")

	return reverse([int(octet) for octet in octets])

def parse_ipv6(addr):
	if addr == "::":
		return [0 for i in range(0, 8)]

	fields = addr.split(":")
	if len(fields) != 8:
		raise RuntimeError("IPv6 address must be in format a:b:c:d:e:f:g:h or '::'")

	return reverse([int(field, 16) for field in fields])

def fmt_ipaddr(addr, ipver):
	if ipver == 4:
		parsed = parse_ipv4(addr)
		ip = struct.pack("<BBBB", *parsed)
		padding = struct.pack("<III", 0, 0, 0)
		return ip + padding
	elif ipver == 6:
		parsed = parse_ipv6(addr)
		return struct.pack("<HHHHHHHH", *parsed)
	else:
		raise RuntimeError("Unsupported IP version: %d" % (ipver,))

def value_of_type(type):
	if type == "ip-and-mask" or type == 0:
		return 0
	elif type == "net-flow" or type == 1:
		return 1
	else:
		raise RuntimeError("Unsupported rule type: %s" % (type))

def fmt_rule(rule):
	type = value_of_type(rule["type"])
	hdr = struct.pack("<BBBB", type, rule["ipver"],
			rule["ipmask"], rule["proto"])

	ip0 = fmt_ipaddr(rule["ip"][0], rule["ipver"])
	ip1 = fmt_ipaddr(rule["ip"][1], rule["ipver"])
	ports = struct.pack("<HH", rule["port"][0], rule["port"][1])

	return hdr + ip0 + ip1 + ports

def fmt_request(req):
	foot = struct.pack("<II", req["rid"], req["sid"])

	if req["action"] == "add-rule" or req["action"] == 0:
		hdr  = struct.pack("<I", 0)
		body = fmt_rule(req["rule"])
		return hdr + body + foot
	elif req["action"] == "del-rule" or req["action"] == 1:
		hdr  = struct.pack("<I", 1)
		body = fmt_rule(req["rule"])
	else:
		raise RuntimeError("Unknown action: %s" % (req["action"],))

	return hdr + body + foot

def parse_ip(addr, ipver):
	if ipver == 4:
		ip = struct.unpack("<BBBB", addr[:4])
		return "%u.%u.%u.%u" % reverse(ip)
	elif ipver == 6:
		ip = struct.unpack("<HHHHHHHH", addr[:16])
		return "%x:%x:%x:%x:%x:%x:%x:%x" % reverse(ip)
	else:
		raise RuntimeError("Unsupported IP version: %u", ipver)

def parse_rule(rule):
	type, ipver, ipmask, proto = struct.unpack("<BBBB", rule[:4])
	ip0 = parse_ip(rule[4:-20], ipver)
	ip1 = parse_ip(rule[-20:-4], ipver)
	ports = struct.unpack("<HH", rule[-4:])

	data = {"type": type, "ipver": ipver,
		"ipmask": ipmask, "proto": proto,
		"ip": [ip0, ip1],
		"port": ports}

	return data

def parse_response(res):
	action, _, status = struct.unpack("<BBh", res[:4])
	data = {"action": action, "status": status}
	footoff = 4

	if action == 0 or action == 1:
		data["rule"] = parse_rule(res[4:44])
		footoff += 40
	else:
		raise RuntimeError("Unknown action: %d" % (action))

	(data["rid"], data["sid"]) = struct.unpack("<II", res[footoff:])
	return data

def do_hello(sock, ccci_ver, id):
	data = struct.pack("<BBI", ccci_ver, 0, id)
	print("Sending: %s" % (binascii.hexlify(data)))
	sock.sendall(data)

	resp = sock.recv(6)
	print("Received: %s" % (binascii.hexlify(resp)))

def do_requests(sock, requests):
	print("Sending %d requests" % (len(requests)))
	for req in requests:
		data = fmt_request(req)
		print("Request: %s (%d B)" % (binascii.hexlify(data), len(data)))
		try:
			print("%s" % (yaml.dump(req)))
		except RuntimeError as e:
			print(e)

		sock.sendall(data)
		res = sock.recv(len(data))

		print("Response: %s (%d B)" % (binascii.hexlify(res), len(res)))
		try:
			print("%s" % (yaml.dump(parse_response(res))))
		except RuntimeError as e:
			print(e)

	time.sleep(10)

def client_service(sock, addr, requests):
	print("Connection from %s" % (addr,))
	do_hello(sock, 1, 1)
	do_requests(sock, requests)
	sock.close()

def server_loop(ssock, requests):
	while 1:
		(csock, addr) = ssock.accept()
		try:
			client_service(csock, addr, requests)
		finally:
			print("Closing client socket")
			csock.close()

def main(argv):
	data = yaml.load_all(sys.stdin)
	requests = []
	for d in data:
		requests.append(d)

	port = len(argv) > 1 and int(argv[1]) or 21105
	ssock = server_create(port, len(argv) > 2 and argv[2] or "0.0.0.0")
	print("Listening on port %d" % (port))

	try:
		server_loop(ssock, requests)
	finally:
		print("Closing server socket")
		ssock.close()

main(sys.argv)
