/**
 * cccid.c
 * Copyright (C) 2012, 2013 Jan Viktorin
 */

#include "ccci_hdr.h"
#include "ccci_msg.h"

#include "cccid_conf.h"
#include "cccid_conn.h"
#include "cccid_sig.h"
#include "cccid_mod_app.h"

#include "version.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>

#define DEFAULT_HOST       "127.0.0.1"
#define DEFAULT_PORT       "21105"
#define DEFAULT_TIMEOUT      0
#define DEFAULT_RETRIES     25
#define DEFAULT_RETRY_DELAY 10
#define DEFAULT_IDENT_FAIL  10
#define DEFAULT_CCCI_VER    0x01

static
int daemon_identify(struct cccid_conf *conf, struct cccid_conn *conn)
{
	const uint8_t ccci_ver = cccid_conf_geti(conf, "ccci.version", DEFAULT_CCCI_VER);
	const uint32_t probeid = cccid_conf_geti(conf, "probe.id", rand());

	union ccci_hdr sendhdr = ccci_hdr_new();
	ccci_hdr_setversion(&sendhdr, ccci_ver);
	ccci_hdr_setprobeid(&sendhdr, probeid);

	syslog(LOG_INFO, "Sending hello by CCCI 0x%" PRIx8 " as device 0x%" PRIx32,
			ccci_ver, probeid);

	if(cccid_conn_send(conn, sendhdr.raw, sizeof(sendhdr.raw)))
		return 1;

	//////////////////////////////////

	union ccci_hdr recvhdr = ccci_hdr_new();
	if(cccid_conn_recv(conn, recvhdr.raw, sizeof(recvhdr.raw)))
		return 2;

	syslog(LOG_INFO, "Received hello by CCCI 0x%" PRIx8 " from 0x%" PRIx32,
			ccci_hdr_getversion(&recvhdr), ccci_hdr_getprobeid(&recvhdr));

	if(ccci_hdr_getversion(&recvhdr) != ccci_ver)
		return 3;

	return 0;
}

static
int daemon_do_service(struct cccid_conf *conf, union ccci_msg *req,
		union ccci_msg *res)
{
	struct cccid_mod *mod = NULL;
	int ret = 0;
	int done = 0;

	do {
		ret = 0;
		mod = cccid_mod_find_matching(mod, req);

		if(mod == NULL && done)
			break;
		else if(mod == NULL) {
			syslog(LOG_WARNING, "No module to service the request");
			return -1;
		}

		syslog(LOG_DEBUG, "Calling mod_%s 0x%" PRIx16, mod->name, mod->version);

		assert(mod->service != NULL);
		ret = mod->service(req, res);

		syslog(LOG_DEBUG, "Result of mod_%s 0x%" PRIx16 ": %d", mod->name, mod->version, ret);

		done = 1;
	} while(ret > 0);

	return ret;
}

static
void daemon_error_response(struct cccid_conf *conf, struct cccid_conn *conn)
{
	union ccci_msg err = ccci_msg_new();

	ccci_msg_setstatus(&err, 1);
	cccid_conn_send(conn, err.raw, sizeof(err.raw));
}

static
void daemon_loop(struct cccid_conf *conf, struct cccid_conn *conn)
{
	do {
		union ccci_msg req = ccci_msg_new();
		union ccci_msg res = ccci_msg_new();

		if(cccid_conn_recv(conn, req.raw, sizeof(req.raw))) {
			if(errno == EINTR || errno == EAGAIN)
				continue;

			syslog(LOG_ERR, "Failed to receive a request (errno %s)", strerror(errno));
			break;
		}

		if(daemon_do_service(conf, &req, &res)) {
			syslog(LOG_ALERT, "Failed to process the request");
			daemon_error_response(conf, conn);
			continue;
		}

		if(cccid_conn_send(conn, res.raw, sizeof(res.raw))) {
			if(errno == EINTR || errno == EAGAIN)
				continue;

			syslog(LOG_ERR, "Failed to send a response (errno %s)", strerror(errno));
			break;
		}

	} while(!cccid_sig_exit());
}

static
void daemon_start(struct cccid_conf *conf)
{
	struct cccid_conn conn = cccid_conn_new();
	int ident_failed = 0;
	const int max_ident_failed = cccid_conf_geti(conf, "ident.failed.max", DEFAULT_IDENT_FAIL);

	do {
		const char *host = cccid_conf_gets(conf, "conn.tcp.host", DEFAULT_HOST);
		const char *port = cccid_conf_gets(conf, "conn.tcp.port", DEFAULT_PORT);

		conn.timeout = cccid_conf_geti(conf, "conn.timeout", DEFAULT_TIMEOUT);
		conn.retries = cccid_conf_geti(conf, "conn.retries", DEFAULT_RETRIES);
		conn.retry_delay = cccid_conf_geti(conf, "conn.retry_delay", DEFAULT_RETRY_DELAY);
		conn.retry_delay = conn.retry_delay < 0? 0 : conn.retry_delay;

		if(ident_failed % max_ident_failed == 0 && ident_failed > 0) {
			syslog(LOG_ALERT, "Failed %d times to identify the server, sleep for %d s",
					ident_failed, conn.retry_delay);
			sleep(conn.retry_delay);
		}

		if(cccid_tcp_open(&conn, host, port)) {
			syslog(LOG_ALERT, "Could not connect to %s at port %s", host, port);
			return;
		}

		if(daemon_identify(conf, &conn)) {
			syslog(LOG_ALERT, "Could not identify the remote host");
			cccid_conn_close(&conn);
			ident_failed += 1;
			continue;
		}

		ident_failed = 0;
		daemon_loop(conf, &conn);
		cccid_conn_close(&conn);
	} while(!cccid_sig_exit());
}

static
void daemon_setlogging(struct cccid_conf *conf)
{
	const char *level = cccid_conf_gets(conf, "log.level", "INFO");
	int mask = 0;

#define MASK_EMERG  (LOG_MASK(LOG_EMERG))
#define MASK_ALERT  (MASK_EMERG  | LOG_MASK(LOG_ALERT))
#define MASK_CRIT   (MASK_ALERT  | LOG_MASK(LOG_CRIT))
#define MASK_ERR    (MASK_CRIT   | LOG_MASK(LOG_ERR))
#define MASK_WARN   (MASK_ERR    | LOG_MASK(LOG_WARNING))
#define MASK_NOTICE (MASK_WARN   | LOG_MASK(LOG_NOTICE))
#define MASK_INFO   (MASK_NOTICE | LOG_MASK(LOG_INFO))
#define MASK_DEBUG  (MASK_INFO   | LOG_MASK(LOG_DEBUG))

	if(!strcmp("DEBUG", level))
		mask = MASK_DEBUG;
	else if(!strcmp("INFO", level))
		mask = MASK_INFO;
	else if(!strcmp("NOTICE", level))
		mask = MASK_NOTICE;
	else if(!strcmp("WARN", level) || !strcmp("WARNING", level))
		mask = MASK_WARN;
	else if(!strcmp("ERR", level) || !strcmp("ERROR", level))
		mask = MASK_ERR;
	else if(!strcmp("CRIT", level))
		mask = MASK_CRIT;
	else if(!strcmp("ALERT", level))
		mask = MASK_ALERT;
	else if(!strcmp("EMERG", level))
		mask = MASK_EMERG;
	else {
		syslog(LOG_WARNING, "Invalid logging level '%s' "
			"(try one of: DEBUG, INFO, NOTICE, WARN, ERR, CRIT, ALERT, EMERG)", level);
		return;
	}

	setlogmask(mask);
}

static
int daemon_help(const char *pname)
{
	fprintf(stderr, "Usage: %s [-H <host>] [-P <port>] [-t <timeout>] ...\n", pname);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -H <host>     connect to the given <host> (default: 127.0.0.1)\n");
	fprintf(stderr, "  -P <port>     connect to the given <port> (default: 5000)\n");
	fprintf(stderr, "  -t <timeout>  timeout for network operations in seconds (default: 0 ~ off)\n");
	fprintf(stderr, "  -r <retries>  number of retries to connect to the remote host after\n");
	fprintf(stderr, "                a failure (default: 10)\n");
	fprintf(stderr, "  -y <delay>    delay between two connection retries in seconds (default: 10)\n");
	fprintf(stderr, "  -v <version>  version of CCCI to expect (default: 1)\n");
	fprintf(stderr, "  -i <probeid>  probe ID to be used (default: random)\n");
	fprintf(stderr, "  -l <level>    log level (default: INFO)\n");
	fprintf(stderr, "  -x <path>     path to the HWIO src\n");
	fprintf(stderr, "  -d <path>     path to the HWIO phys\n");
	fprintf(stderr, "  -h, --help    print help\n");
	return 0;
}

int main(int argc, char **argv)
{
	cccid_sig_install();
	openlog(argv[0], LOG_NDELAY | LOG_PID | LOG_PERROR, LOG_DAEMON);
	srand(time(NULL));

	for(int i = 1; i < argc; ++i) {
		if(!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
			return daemon_help(argv[0]);
	}

	struct cccid_conf conf = cccid_conf_new();
	if(cccid_conf_argv(&conf, argc, argv))
		goto exit_gracefully;

	syslog(LOG_NOTICE, "Starting cccid %s (built %s %s)", CCCID_VERSION, __DATE__, __TIME__);
	daemon_setlogging(&conf);

	cccid_mod_install_all(&conf);

	daemon_start(&conf);

	cccid_mod_uninstall_all();

exit_gracefully:
	cccid_conf_fini(&conf);
	closelog();
	return 0;
}
