/**
 * cccid_conf.c
 * Copyright (C) 2012, 2013 Jan Viktorin
 */

#define _XOPEN_SOURCE

#include "cccid_conf.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

static
void *memdup(const void *m, size_t len)
{
	if(len == 0)
		return NULL;

	void *dup = malloc(len);
	if(dup == NULL)
		return NULL;

	memcpy(dup, m, len);
	return dup;
}

/**
 * Set the given string s to the new value news.
 * If the orignal string s is already allocated,
 * it is reallocated if necessary.
 */
static
int string_set(char **s, const char *news)
{
	assert(s != NULL);
	assert(news != NULL);

	const size_t newlen = strlen(news);

	if(*s == NULL) {
		*s = (char *) memdup(news, newlen + 1);
		return *s == NULL;
	}

	const size_t oldlen = strlen(*s);

	if(oldlen >= newlen) {
		memcpy(*s, news, newlen + 1);
		return 0;
	}

	void *m = realloc(*s, newlen + 1);
	if(m == NULL)
		return -1;

	memcpy(m, news, newlen + 1);
	*s = m;
	return 0;
}

/**
 * Set value of an existing key at the index i.
 * Private function, see conf_sets.
 */
static
int sets_at(struct cccid_conf *conf, int i, const char *value)
{
	if(value == NULL) {
		if(conf->values[i] == NULL)
			return 0;

		free(conf->values[i]);
		conf->values[i] = NULL;
		return 0;
	}

	return string_set(&conf->values[i], value);
}

/**
 * Insert a new pair key, value at the index i.
 * Private function, see conf_sets.
 */
static
int insert_and_sets(struct cccid_conf *conf, int i, const char *key,
		const char *value)
{
	if(string_set(&conf->keys[i], key))
		return 1;

	return sets_at(conf, i, value);
}

/**
 * Reallocates the configuration arrays and inserts the new
 * pair key, value.
 * Private function, see conf_sets.
 */
static
int realloc_and_sets(struct cccid_conf *conf, const char *key,
		const char *value)
{
	const int newcap = conf->cap + 16;

	char **keys = (char **) realloc(conf->keys, newcap * sizeof(char *));
	if(keys == NULL)
		return -1;
	else
		conf->keys = keys;

	char **values = (char **) realloc(conf->values, newcap * sizeof(char *));
	if(values == NULL)
		return -1;
	else
		conf->values = values;

	const int oldcap = conf->cap;
	conf->cap = newcap;

	memset(conf->keys + oldcap, 0, (newcap - oldcap) * sizeof(char *));
	memset(conf->values + oldcap, 0, (newcap - oldcap) * sizeof(char *));

	return insert_and_sets(conf, oldcap, key, value);
}

/**
 * Sets value of the given key.
 * An internal API function.
 */
static
int conf_sets(struct cccid_conf *conf, const char *key,
		const char *value)
{
	int target = -1;

	for(int i = 0; i < conf->cap; ++i) {
		if(target < 0 && conf->keys[i] == NULL)
			target = i;

		if(conf->keys[i] != NULL && !strcmp(key, conf->keys[i])) {
			target = i;
			break;
		}
	}

	if(target < 0)
		return realloc_and_sets(conf, key, value);
	else if(conf->keys[target])
		return sets_at(conf, target, value);
	else
		return insert_and_sets(conf, target, key, value);
}

void cccid_conf_fini(struct cccid_conf *conf)
{
	for(size_t i = 0; i < conf->cap; ++i) {
		if(conf->keys && conf->keys[i])
			free(conf->keys[i]);

		if(conf->values && conf->values[i])
			free(conf->values[i]);
	}

	if(conf->keys)
		free(conf->keys);
	if(conf->values)
		free(conf->values);

	memset(conf, 0, sizeof(struct cccid_conf));
}

int cccid_conf_argv(struct cccid_conf *conf, int argc, char **argv)
{
	int opt;

	while((opt = getopt(argc, argv, "H:P:t:r:y:v:i:l:f:d:x:")) != -1) {
		int err = 0;

		switch(opt) {
		case 'H':
			err = conf_sets(conf, "conn.tcp.host", optarg);
			break;
		case 'P':
			err = conf_sets(conf, "conn.tcp.port", optarg);
			break;
		case 't':
			err = conf_sets(conf, "conn.timeout", optarg);
			break;
		case 'r':
			err = conf_sets(conf, "conn.retries", optarg);
			break;
		case 'y':
			err = conf_sets(conf, "conn.retry_delay", optarg);
			break;
		case 'v':
			err = conf_sets(conf, "ccci.version", optarg);
			break;
		case 'i':
			err = conf_sets(conf, "probe.id", optarg);
			break;
		case 'l':
			err = conf_sets(conf, "log.level", optarg);
			break;
		case 'f':
			err = conf_sets(conf, "ident.failed.max", optarg);
			break;
		case 'd':
			err = conf_sets(conf, "mod.hwio.phys", optarg);
			break;
		case 'x':
			err = conf_sets(conf, "mod.hwio.src", optarg);
			break;
		default:
			continue;
		}

		if(err)
			return err;
	}

	return 0;
}

/**
 * Get the value of the given key. If the value does not exist,
 * returns NULL.
 * Internal API function.
 */
static
const char *conf_gets(const struct cccid_conf *conf, const char *key)
{
	assert(key != NULL);

	for(int i = 0; i < conf->cap; ++i) {
		if(conf->keys[i] && !strcmp(conf->keys[i], key))
			return conf->values[i]? conf->values[i] : NULL;
	}

	return NULL;
}

const char *cccid_conf_gets(struct cccid_conf *conf, const char *key,
		const char *defval)
{
	const char *s = conf_gets(conf, key);
	return s? s : defval;
}

int cccid_conf_geti(struct cccid_conf *conf, const char *key,
		int defval)
{
	const char *s = conf_gets(conf, key);
	if(s == NULL)
		return defval;

	return atoi(s);
}

#ifdef CCCID_CONF_TEST
#include "cccid_test.h"

static
void test_memdup(void)
{
	char *source = "testing string";

	void *a = memdup(source, 0);
	assert(a == NULL);
	test_ok("Memdup of 0 bytes");

	////////////////////////////////////

	void *b = memdup(source, 1);
	assert(b != NULL);
	assert(b != source);
	assert(!memcmp(source, b, 1));
	free(b);
	test_ok("Memdup of 1 byte");

	////////////////////////////////////

	void *c = memdup(source, strlen(source));
	assert(c != NULL);
	assert(c != source);
	assert(!memcmp(source, c, strlen(source)));
	free(c);
	test_ok("Memdup of %zu bytes", strlen(source));

	////////////////////////////////////

	void *d = memdup(source, strlen(source) + 1);
	assert(d != NULL);
	assert(d != source);
	assert(!strcmp(source, (char *) d));
	free(d);
	test_ok("Memdup of %zu bytes", strlen(source) + 1);
}

static
void test_string_set(void)
{
	const char *news = "my new string";
	const char *news_long = "my new very long string";
	char *s = NULL;
	const char *orig_s = NULL;

	if(string_set(&s, news))
		test_failed("Failed to set %p to the new string %s", &s, news);
	assert(s != NULL);
	assert(!strcmp(s, news));
	free(s);
	test_ok("Set non-allocated string");

	////////////////////////////////////

	s = memdup(news, strlen(news) + 1);
	orig_s = s;
	if(string_set(&s, news))
		test_failed("Failed to set %p to the same string %s", &s, news);
	assert(s != NULL);
	assert(s == orig_s);
	assert(!strcmp(s, news));
	assert(s != news);
	free(s);
	test_ok("Overwrite allocated string");

	////////////////////////////////////

	s = memdup(news, strlen(news) + 1);
	orig_s = s;
	if(string_set(&s, news_long))
		test_failed("Failed to set %p to the longer string %s", &s, news_long);
	assert(s != NULL);
	assert(!strcmp(s, news_long));
	assert(s != news_long);
	free(s);
	test_ok("Overwrite by longer string");

	////////////////////////////////////

	s = memdup(news, strlen(news) + 1);
	orig_s = s;
	if(string_set(&s, ""))
		test_failed("Failed to set %p to empty string", &s);
	assert(s != NULL);
	assert(s == orig_s);
	assert(s[0] == '\0');
	free(s);
	test_ok("Overwrite by empty string");
}

void test_realloc_and_sets(void)
{
	struct cccid_conf conf = cccid_conf_new();

	assert(conf.keys == NULL);
	assert(conf.values == NULL);
	assert(conf.cap == 0);

	////////////////////////////////////

	if(realloc_and_sets(&conf, "key1", "value1"))
		test_failed("Failed to realloc the new config and add ('key1', 'value1')");

	assert(conf.keys != NULL);
	assert(conf.values != NULL);
	assert(conf.cap > 0);

	assert(!strcmp(conf.keys[0], "key1"));
	assert(!strcmp(conf.values[0], "value1"));

	for(int i = 1; i < conf.cap; ++i) {
		assert(conf.keys[i] == NULL);
		assert(conf.values[i] == NULL);
	}

	test_ok("Successfully reallocated and added ('key1', 'value1')");

	////////////////////////////////////

	const int oldcap = conf.cap;

	for(int i = 1; i < conf.cap; ++i) {
		char key[1024];
		sprintf(key, "key%d", i);

		if(insert_and_sets(&conf, i, key, "value1"))
			test_failed("Failed to insert key '%s'", key);

		assert(conf.keys[i] != NULL);
		assert(conf.values[i] != NULL);
		assert(!strcmp(conf.keys[i], key));
		assert(!strcmp(conf.values[i], "value1"));

		assert(conf.cap == oldcap);
	}

	test_ok("Successfully reached the capacity of the cccid_conf");

	////////////////////////////////////

	assert(conf.cap == oldcap);

	if(realloc_and_sets(&conf, "magic", "magic_value"))
		test_failed("Failed to insert over capacity ('magic', 'magic_value')");

	assert(conf.keys != NULL);
	assert(conf.values != NULL);
	assert(conf.cap > oldcap);

	assert(!strcmp(conf.keys[oldcap], "magic"));
	assert(!strcmp(conf.values[oldcap], "magic_value"));

	for(int i = oldcap + 1; i < conf.cap; ++i) {
		assert(conf.keys[i] == NULL);
		assert(conf.values[i] == NULL);
	}

	test_ok("Successfully reallocated and added ('magic', 'magic_value')");

	////////////////////////////////////

	cccid_conf_fini(&conf);
}

void test_conf_sets(void)
{
	struct cccid_conf conf = cccid_conf_new();

	if(conf_sets(&conf, "key1", "value1"))
		test_failed("Failed to set ('key1', 'value1')");
	assert(!strcmp(conf.keys[0], "key1"));
	assert(!strcmp(conf.values[0], "value1"));
	test_ok("Successfully set ('key1', 'value1')");

	////////////////////////////////////

	if(conf_sets(&conf, "key2", "value2"))
		test_failed("Failed to set ('key2', 'value2')");
	assert(!strcmp(conf.keys[0], "key1"));
	assert(!strcmp(conf.values[0], "value1"));
	assert(!strcmp(conf.keys[1], "key2"));
	assert(!strcmp(conf.values[1], "value2"));
	test_ok("Successfully set ('key2', 'value2')");

	////////////////////////////////////

	if(conf_sets(&conf, "key2", "value3"))
		test_failed("Failed to set ('key2', 'value3')");
	assert(!strcmp(conf.keys[0], "key1"));
	assert(!strcmp(conf.values[0], "value1"));
	assert(!strcmp(conf.keys[1], "key2"));
	assert(!strcmp(conf.values[1], "value3"));
	test_ok("Successfully set ('key2', 'value3')");

	////////////////////////////////////

	if(conf_sets(&conf, "key2", NULL))
		test_failed("Failed to set ('key2', NULL)");
	assert(!strcmp(conf.keys[0], "key1"));
	assert(!strcmp(conf.values[0], "value1"));
	assert(!strcmp(conf.keys[1], "key2"));
	assert(conf.values[1] == NULL);
	test_ok("Successfully set ('key2', NULL)");

	////////////////////////////////////

	if(conf_sets(&conf, "key3", NULL))
		test_failed("Failed to set ('key3', NULL)");
	assert(!strcmp(conf.keys[0], "key1"));
	assert(!strcmp(conf.values[0], "value1"));
	assert(!strcmp(conf.keys[1], "key2"));
	assert(conf.values[1] == NULL);
	assert(!strcmp(conf.keys[2], "key3"));
	assert(conf.values[2] == NULL);
	test_ok("Successfully set ('key3', NULL)");

	cccid_conf_fini(&conf);
}

static
void test_conf_gets(void)
{
	struct cccid_conf conf = cccid_conf_new();

	if(conf_sets(&conf, "key1", "value1"))
		test_failed("Failed to insert ('key1', 'value1')");
	if(conf_sets(&conf, "key2", "value2"))
		test_failed("Failed to insert ('key2', 'value2')");
	if(conf_sets(&conf, "key3", "value3"))
		test_failed("Failed to insert ('key3', 'value3')");

	if(conf_gets(&conf, "key1") == NULL)
		test_failed("Can not find 'key1'");
	if(strcmp(conf_gets(&conf, "key1"), "value1"))
		test_failed("Invalid value for key 'key1': %s", conf_gets(&conf, "key1"));

	test_ok("Key 'key1' found");

	if(conf_gets(&conf, "key2") == NULL)
		test_failed("Can not find 'key2'");
	if(strcmp(conf_gets(&conf, "key2"), "value2"))
		test_failed("Invalid value for key 'key2': %s", conf_gets(&conf, "key2"));

	test_ok("Key 'key2' found");

	if(conf_gets(&conf, "key3") == NULL)
		test_failed("Can not find 'key3'");
	if(strcmp(conf_gets(&conf, "key3"), "value3"))
		test_failed("Invalid value for key 'key3': %s", conf_gets(&conf, "key3"));

	test_ok("Key 'key3' found");

	if(conf_gets(&conf, "key4") != NULL)
		test_failed("The 'key4' was found, but must not be");

	test_ok("Key 'key4' was not found");

	if(conf_gets(&conf, "key5") != NULL)
		test_failed("The 'key5' was found, but must not be");

	test_ok("Key 'key5' was not found");

	cccid_conf_fini(&conf);
}

int main(int argc, char **argv)
{
	test_memdup();
	test_string_set();
	test_realloc_and_sets();
	test_conf_sets();
	test_conf_gets();
	return 0;
}
#endif
