/**
 * cccid_conn.c
 * Copyright (C) 2012, 2013 Jan Viktorin
 */

#include "cccid_conn.h"
#include <unistd.h>
#include <errno.h>
#include <syslog.h>

void cccid_conn_close(struct cccid_conn *conn)
{
	if(conn->fd >= 0) {
		close(conn->fd);
		conn->fd = -1;
	}
}

int cccid_conn_send(struct cccid_conn *conn,
		const uint8_t *b, size_t blen)
{
	syslog(LOG_DEBUG, "Sending %zu B", blen);

	size_t total = 0;

	do {
		ssize_t wlen = write(conn->fd, b + total, blen - total);
		if(wlen < 0 && (errno == EINTR || errno == EAGAIN))
			return 1;
		else if(wlen == 0)
			break;
		else if(wlen < 0)
			return -1;
		else
			total += wlen;
	} while(total < blen);

	return total != blen;
}

int cccid_conn_recv(struct cccid_conn *conn,
		uint8_t *b, size_t blen)
{
	syslog(LOG_DEBUG, "Receving %zu B", blen);

	size_t total = 0;

	do {
		ssize_t rlen = read(conn->fd, b + total, blen - total);
		if(rlen < 0 && (errno == EINTR || errno == EAGAIN))
			return 1;
		else if(rlen == 0)
			break;
		else if(rlen < 0)
			return -1;
		else 
			total += rlen;
	} while(total < blen);

	return total != blen;
}

#ifdef CCCID_CONN_TEST
#include "cccid_test.h"

void test_send_recv(void)
{
	int pipes[2];

	if(pipe(pipes) < 0)
		test_failed("Failed to create pipes: %s", strerror(errno));

	struct cccid_conn conn = cccid_conn_new();
	uint8_t buff[1024];

	switch(fork()) {
	case 0:
		conn.fd = pipes[0];

		for(int j = 0; j < 8; ++j) {
			if(cccid_conn_recv(&conn, buff, sizeof(buff)))
				test_failed("Failed to receive data: %s", strerror(errno));

			for(int i = 0; i < sizeof(buff); ++i) {
				if(buff[i] != (uint8_t) (i * j))
					test_failed("Invalid byte at %d: 0x%x, expect 0x%x", i, buff[i], (uint8_t) (i * j));
			}
		}

		test_ok("Received data");
		cccid_conn_close(&conn);
		exit(0);

	case -1:
		test_failed("Failed to fork: %s", strerror(errno));

	default:
		conn.fd = pipes[1];

		for(int j = 0; j < 8; ++j) {
			for(int i = 0; i < sizeof(buff); ++i)
				buff[i] = i * j;

			if(cccid_conn_send(&conn, buff, sizeof(buff)))
				test_failed("Failed to send data: %s", strerror(errno));
		}

		test_ok("Sent data");
		cccid_conn_close(&conn);
		break;
	}
}

int main(int argc, char **argv)
{
	test_send_recv();
	return 0;
}

#endif
