/**
 * cccid_tcp.c
 * Copyright (C) 2012, 2013 Jan Viktorin
 */

#define _POSIX_SOURCE

#include "cccid_conn.h"
#include "cccid_sig.h"

#include <syslog.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#define max(a, b) ((a) > (b)? (a) : (b))

static
struct addrinfo *ask_host_and_service(const char *host, const char *service)
{
	static const struct addrinfo hints = {
		.ai_family = AF_UNSPEC,
		.ai_socktype = SOCK_STREAM,
		.ai_protocol = 0,
		.ai_flags = AI_CANONNAME,
		.ai_addrlen = 0,
		.ai_addr = NULL,
		.ai_canonname = NULL,
		.ai_next = NULL
	};

	struct addrinfo *ainfo = NULL;

	int err = getaddrinfo(host, service, &hints, &ainfo);
	if(err) {
		syslog(LOG_ERR, "getaddrinfo(): %s", gai_strerror(err));
		return NULL;
	}

	return ainfo;
}

static
const char *sockaddr_family_tostr(const struct sockaddr *addr)
{
	assert(addr != NULL);

	if(addr->sa_family == AF_INET)
		return "AF_INET";

	if(addr->sa_family == AF_INET6)
		return "AF_INET6";

	return "AF_???";
}

static
const char *sockaddr_addr_tostr(const struct sockaddr *addr)
{
	assert(addr != NULL);

	static char buff[max(INET_ADDRSTRLEN, INET6_ADDRSTRLEN)];

	if(addr->sa_family == AF_INET) {
		return inet_ntop(AF_INET,
				&((struct sockaddr_in *) addr)->sin_addr,
				buff, sizeof(buff));
	}

	if(addr->sa_family == AF_INET6) {
		return inet_ntop(AF_INET6,
				&((struct sockaddr_in6 *) addr)->sin6_addr,
				buff, sizeof(buff));
	}

	return "(unknown)";
}

static
uint16_t sockaddr_port(const struct sockaddr *addr)
{
	assert(addr != NULL);

	if(addr->sa_family == AF_INET)
		return ntohs(((struct sockaddr_in *) addr)->sin_port);

	if(addr->sa_family == AF_INET6)
		return ntohs(((struct sockaddr_in6 *) addr)->sin6_port);

	return 0;
}

static
int sock_set_timeout(int sock, int timeout)
{
	if(timeout <= 0)
		return 0;

	struct timeval t = {
		.tv_sec = timeout,
		.tv_usec = 0
	};

	if(setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &t, sizeof(t)) < 0) {
		syslog(LOG_ERR, "setsockopt(SO_RCVTIMEO): %s", strerror(errno));
		return -1;
	}

	if(setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, &t, sizeof(t)) < 0) {
		syslog(LOG_ERR, "setsockopt(SO_SNDTIMEO): %s", strerror(errno));
		return -2;
	}

	return 0;
}

static
int socket_and_connect(const struct sockaddr *addr, socklen_t addrlen, int timeout)
{
	int sock = socket(addr->sa_family, SOCK_STREAM, 0);

	if(sock < 0) {
		syslog(LOG_ERR, "socket(%s): %s", sockaddr_family_tostr(addr), strerror(errno));
		return -1;
	}

	if(sock_set_timeout(sock, timeout)) {
		close(sock);
		return -2;
	}

	if(connect(sock, addr, addrlen) < 0) {
		syslog(LOG_ERR, "connect(%s:%" PRIu16 "): %s", sockaddr_addr_tostr(addr),
				sockaddr_port(addr), strerror(errno));
		close(sock);
		return -3;
	}

	return sock;
}

static
int sock_open_and_connect(struct cccid_conn *conn, const char *host, const char *service)
{
	struct addrinfo *ainfo = ask_host_and_service(host, service);
	struct addrinfo *caddr;

	for(caddr = ainfo; caddr != NULL; caddr = caddr->ai_next) {
		if(caddr->ai_family != AF_INET && caddr->ai_family != AF_INET6)
			continue; // just IPv4 or IPv6 is allowed

		struct sockaddr *addr = (struct sockaddr *) caddr->ai_addr;
		int sock = socket_and_connect(addr, caddr->ai_addrlen, conn->timeout);

		if(sock < 0)
			continue;

		syslog(LOG_INFO, "Connected to %s (%s:%" PRIu16 ")", ainfo->ai_canonname,
				sockaddr_addr_tostr(caddr->ai_addr), sockaddr_port(caddr->ai_addr));

		freeaddrinfo(ainfo);
		conn->fd = sock;
		return 0;
	}

	if(ainfo != NULL)
		freeaddrinfo(ainfo);

	return 2;
}

int cccid_tcp_open(struct cccid_conn *conn, const char *host,
		const char *service)
{
	assert(conn->fd < 0);

	for(int i = 0; conn->retries < 0 || i <= conn->retries; ++i) {
		if(conn->fd >= 0)
			close(conn->fd);

		syslog(LOG_DEBUG, "Attempt to connect to %s:%s", host, service);

		if(!sock_open_and_connect(conn, host, service))
			return 0;

		if(cccid_sig_exit())
			break;

		syslog(LOG_ALERT, "Failed to connect to %s:%s, retry in %d s (%d/%d)",
				host, service, conn->retry_delay, i, conn->retries);

		if(conn->retry_delay > 0)
			sleep(conn->retry_delay);
	}

	close(conn->fd);
	return -3;
}
