/**
 * hwio_filter.c
 * Copyright (C) 2012, 2013 Jan Viktorin
 */

#define _BSD_SOURCE

#include "../compat_endian.h"
#include "hwio_filter.h"
#include "rid_mgmt.h"

#include <inttypes.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>

#ifndef RULES_MAX
#define RULES_MAX 1024
#endif

#define REGW_SRCIP  0x024
#define REGW_DATA   0x018
#define REGW_ADDCMD 0x008
#define REGW_DELCMD 0x00C

#define REGR_FULL   0x008
#define REGR_BUSY   0x00C

#define RULE_TYPE_IP24  0x001
#define RULE_TYPE_IP48  0x002
#define RULE_TYPE_IP64  0x004
#define RULE_TYPE_IP128 0x008
#define RULE_INVALID    0x000

#define MASK_GLOBAL_BUSY 0x001

///////////////////////////////////

static
int loop_on_status(struct hwio_comp *comp, const uint32_t regaddr,
		const uint32_t mask, const int loops_limit,
		const int delay)
{
	uint32_t status = 0;

	if(hwio_comp_read(comp, regaddr, &status))
		return -1;

	for(int i = 0; status & mask; ++i) {
		if(i >= loops_limit)
			return 1;
		else if(i > 0)
			usleep(delay);

		if(hwio_comp_read(comp, regaddr, &status))
			return -1;
	}

	return 0;
}

static
int wait_until_not_busy(struct hwio_comp *filter)
{
	return loop_on_status(filter, REGR_BUSY,
			 MASK_GLOBAL_BUSY, 1024, 1);
}

static
int filter_is_full(struct hwio_comp *filter, uint32_t rule_type)
{
	return loop_on_status(filter, REGR_FULL,
			rule_type, 1024, 100);
}

static
int ipv4_prepare(struct hwio_comp *comp, const union ccci_ipaddr *addr)
{
	for(int i = 0; i < 3; ++i) {
		if(hwio_comp_write(comp, REGW_SRCIP + i * 4, 0))
			return 1;
	}

	if(hwio_comp_write(comp, REGW_SRCIP + 0xC, native_to_be(ccci_ipaddr_getipv4(addr))))
		return 1;

	return 0;
}

static
int ipv6_prepare(struct hwio_comp *comp, const union ccci_ipaddr *addr)
{
	uint32_t value[4];
	ccci_ipaddr_get32_ipv6(addr, value);

	for(int i = 0; i < 4; ++i) {
		if(hwio_comp_write(comp, REGW_SRCIP + i * 4, native_to_be(value[3 - i])))
			return 1;
	}

	return 0;
}

static
int rid_prepare(struct hwio_comp *comp, rid_t rid)
{
	return hwio_comp_write(comp, REGW_DATA, native_to_le(rid));
}

static
int rule_prepare(struct hwio_comp *comp, uint32_t rule_type, rid_t rid,
		const union ccci_rule *rule)
{
	const uint8_t ipver = ccci_rule_getipver(rule);
	const union ccci_ipaddr *addr = ccci_rule_getip0(rule);

	switch(rule_type) {
	case RULE_TYPE_IP24:
	case RULE_TYPE_IP48:
	case RULE_TYPE_IP64:
	case RULE_TYPE_IP128:
		if(ipver == 4) {
			if(ipv4_prepare(comp, addr))
				return 1;
		}
		else if(ipver == 6) {
			if(ipv6_prepare(comp, addr))
				return 1;
		}

		return rid_prepare(comp, rid);

	default:
		return -1;
	}
}

static
int rule_commit(struct hwio_comp *comp, uint32_t rule_mask)
{
	return hwio_comp_write(comp, REGW_ADDCMD, rule_mask);
}

static
int rule_delete(struct hwio_comp *comp, uint32_t rule_mask)
{
	return hwio_comp_write(comp, REGW_DELCMD, rule_mask);
}

static
uint32_t rule_type_from_ipmask(uint8_t ipmask, uint8_t ipver)
{
	assert(ipver == 4 || ipver == 6);

	if(ipver == 4) {
		if(ipmask == 24)
			return RULE_TYPE_IP24;

		if(ipmask == 32)
			return RULE_TYPE_IP128;
	}
	else {/* ipver == 6 */
		if(ipmask == 48)
			return RULE_TYPE_IP48;

		if(ipmask == 64)
			return RULE_TYPE_IP64;

		if(ipmask == 128)
			return RULE_TYPE_IP128;
	}

	return RULE_INVALID;
}

///////////////////////////////////

/**
 * Filter to be accessed.
 */
static struct hwio_comp *comp_filter;

/**
 * Management of allocated RIDs.
 */
static struct rid_mgmt rid_mgmt;

///////////////////////////////////

int hwio_filter_init(struct hwio_comp *filter)
{
	assert(filter != NULL);
	assert(comp_filter == NULL);

	if(rid_mgmt_init(&rid_mgmt, RULES_MAX)) {
		syslog(LOG_ALERT, "Failed to init the RID manager for %u rules", RULES_MAX);
		return 1;
	}

	comp_filter = filter;
	return 0;
}

void hwio_filter_fini(void)
{
	assert(comp_filter != NULL);

	rid_mgmt_fini(&rid_mgmt);

	hwio_comp_free(comp_filter);
	comp_filter = NULL;
}

uint16_t hwio_filter_addrule(const union ccci_rule *rule,
		uint32_t sid, uint32_t *rid32)
{
	rid_t rid;
	uint32_t rule_type = RULE_INVALID;

	switch(ccci_rule_gettype(rule)) {
	case 0: // ip and ipmask
		rule_type = rule_type_from_ipmask(ccci_rule_getipmask(rule),
					ccci_rule_getipver(rule));
		if(rule_type == RULE_INVALID)
			return 0;

		if(rid_mgmt_alloc(&rid_mgmt, &rid))
			return 0;

		if(wait_until_not_busy(comp_filter)) {
			rid_mgmt_release(&rid_mgmt, rid);
			return 0;
		}

		if(filter_is_full(comp_filter, rule_type)) {
			rid_mgmt_release(&rid_mgmt, rid);
			return 0;
		}

		if(rule_prepare(comp_filter, rule_type, rid, rule)) {
			rid_mgmt_release(&rid_mgmt, rid);
			return 0;
		}

		if(rule_commit(comp_filter, rule_type)) {
			rid_mgmt_release(&rid_mgmt, rid);
			return 0;
		}

		*rid32 = (uint32_t) rid;
		break;

	case 1: // unsupported
		return 0;

	case 2: // unsupported
		return 0;
	case 3: // unsupported
		return 0;
	default: // invalid
		return 0;
	}

	return 1;
}

uint16_t hwio_filter_delrule(const union ccci_rule *rule,
		uint32_t sid, uint32_t rid)
{
	uint32_t rule_type = RULE_INVALID;

	if(ccci_rule_gettype(rule) == 0) {
		rule_type = rule_type_from_ipmask(ccci_rule_getipmask(rule),
					ccci_rule_getipver(rule));
		if(rule_type == RULE_INVALID)
			return 0;

		if(rule_prepare(comp_filter, rule_type, (rid_t) rid, rule))
			return 0;

		if(rule_delete(comp_filter, rule_type))
			return 0;

		rid_mgmt_release_later(&rid_mgmt, rid);
		return 1;
	}

	return 0;
}
