/**
 * mod_hwio_filter.c
 * Copyright (C) 2012, 2013 Jan Viktorin
 */

#include "../ccci_msg.h"
#include "../cccid_mod.h"

#include "hwio.h"
#include "hwio_filter.h"

#include <inttypes.h>
#include <assert.h>
#include <string.h>

static
int filter_match(const union ccci_msg *msg)
{
	const uint8_t action = ccci_msg_getaction(msg);
	return action != 0 && action != 1;
}

static
int filter_service(const union ccci_msg *req, union ccci_msg *res)
{
	uint16_t status = 0;
	uint32_t rid;

	memcpy(res, req, sizeof(union ccci_msg));

	switch(ccci_msg_getaction(req)) {
	case 0:
		status = hwio_filter_addrule(ccci_msg_getrule(req),
				ccci_msg_getsid(req), &rid);
		ccci_msg_setrid(res, rid);
		break;

	case 1:
		status = hwio_filter_delrule(ccci_msg_getrule(req),
				ccci_msg_getsid(req), ccci_msg_getrid(req));
		break;

	default:
		return -1;
	}

	ccci_msg_setstatus(res, status);
	return 1;
}

///////////////////////////////////

/**
 * HWIO Device instance.
 */
static
struct hwio_dev dev;

/**
 * List of compatible filters.
 */
static
struct hwio_comp_spec filter_compat[] = {
	{"xlnx",    "axi_advanced_filter", HWIO_VER3(1, 0, HWIO_VERSION_NA)},
	{"combov2", "Advanced_filter",     HWIO_VER2(1, 0)},
	{"debug",   "Advanced_filter",     HWIO_VER2(1, 0)},
	HWIO_COMPAT_END
};

///////////////////////////////////

static
void filter_fini(void)
{
	hwio_filter_fini();
	hwio_deinit(&dev);
}

static
struct cccid_mod mod_filter = {
	.name = "filter",
	.version = 0x0001,
	.match = filter_match,
	.service = filter_service,
	.fini = filter_fini
};

static
void filter_init(struct cccid_conf *conf)
{
	const char *phys = cccid_conf_gets(conf, "mod.hwio.phys", hwio_default_dev_phys());
	const char *src  = cccid_conf_gets(conf, "mod.hwio.src",  hwio_default_dev_src());

	if(hwio_init(&dev, phys, src)) {
		syslog(LOG_ALERT, "Failed to init hwio for the mod_%s 0x%" PRIx16,
				mod_filter.name, mod_filter.version);
		return;
	}

	struct hwio_comp *head = NULL;

	if(hwio_byspec(&dev, filter_compat, &head)) {
		syslog(LOG_ALERT, "Failed to lookup compatible filters");
		hwio_deinit(&dev);
		return;
	}

	if(head == NULL) {
		syslog(LOG_ALERT, "No such compatible filter was found");
		hwio_deinit(&dev);
		return;
	}

	hwio_comp_list_free(hwio_comp_next(head));

	if(hwio_filter_init(head)) {
		syslog(LOG_ALERT, "Failed to allocate the hwio-filter module");
		hwio_comp_free(head);
		hwio_deinit(&dev);
		return;
	}

	if(cccid_mod_install(&mod_filter)) {
		syslog(LOG_ALERT, "Failed to install the mod_%s 0x%" PRIx16,
				mod_filter.name, mod_filter.version);

		hwio_filter_fini();
		hwio_deinit(&dev);
		return;
	}

	syslog(LOG_INFO, "Driving filter %s by mod_%s 0x%" PRIx16,
			hwio_comp_name(head), mod_filter.name, mod_filter.version);
}

cccid_mod_init(hwio_filter, filter_init)
