/**
 * rid_mgmt.c
 * Copyright (C) 2012, 2013 Jan Viktorin
 */

#include "rid_mgmt.h"
#include <stdlib.h>
#include <inttypes.h>
#include <assert.h>

static
int queue_full(const struct rid_mgmt *mgmt)
{
	return ((mgmt->queue_begin + 1) % mgmt->queue_depth) == mgmt->queue_end;
}

static
int queue_empty(const struct rid_mgmt *mgmt)
{
	return mgmt->queue_begin == mgmt->queue_end;
}

static
rid_t queue_take(struct rid_mgmt *mgmt)
{
	assert(!queue_empty(mgmt));

	const rid_t value = mgmt->queue[mgmt->queue_end];

	mgmt->queue_end += 1;
	mgmt->queue_end %= mgmt->queue_depth;

	return value;
}

static
void queue_push(struct rid_mgmt *mgmt, rid_t rid)
{
	assert(!queue_full(mgmt));

	mgmt->queue[mgmt->queue_begin] = rid;

	mgmt->queue_begin += 1;
	mgmt->queue_begin %= mgmt->queue_depth;
}

static
void queue_push_unique(struct rid_mgmt *mgmt, rid_t rid)
{
	size_t i = mgmt->queue_end;

	while(i != mgmt->queue_begin) {
		if(mgmt->queue[i] == rid)
			return;

		i += 1;
		i %= mgmt->queue_depth;
	}

	queue_push(mgmt, rid);
}

////////////////////////////////////////////

int rid_mgmt_init(struct rid_mgmt *mgmt, size_t max)
{
	static const uint32_t bits_left[] = {
		0xFFFFFFFF, 0xFFFFFFFE, 0xFFFFFFFC, 0xFFFFFFF8,
		0xFFFFFFF0, 0xFFFFFFE0, 0xFFFFFFC0, 0xFFFFFF80,
		0xFFFFFF00, 0xFFFFFE00, 0xFFFFFC00, 0xFFFFF800,
		0xFFFFF000, 0xFFFFE000, 0xFFFFC000, 0xFFFF8000,
		0xFFFF0000, 0xFFFE0000, 0xFFFC0000, 0xFFF80000,
		0xFFF00000, 0xFFE00000, 0xFFC00000, 0xFF800000,
		0xFF000000, 0xFE000000, 0xFC000000, 0xF8000000,
		0xF0000000, 0xE0000000, 0xC0000000, 0x80000000
	};
	assert(sizeof(bits_left) / sizeof(bits_left[0]) == 32);

	void *m = calloc(max / 32 + (max % 32? 1 : 0), sizeof(uint32_t));
	if(m == NULL)
		return 1;

	assert(max > 4); // assume that max is quite big

	mgmt->queue = (rid_t *) calloc(max / 4 + 1, sizeof(rid_t));
	if(mgmt->queue == NULL) {
		free(m);
		return 2;
	}

	mgmt->bitmap = m;
	mgmt->max = max;
	mgmt->queue_depth = max / 4 + 1;
	mgmt->queue_begin = 0;
	mgmt->queue_end = 0;

	if(max % 32 != 0) {
		uint32_t *bitmap = (uint32_t *) mgmt->bitmap;
		bitmap[mgmt->max / 32] = bits_left[max % 32];
	}

	return 0;
}

void rid_mgmt_fini(struct rid_mgmt *mgmt)
{
	if(mgmt->bitmap)
		free(mgmt->bitmap);

	if(mgmt->queue)
		free(mgmt->queue);

	mgmt->bitmap = NULL;
	mgmt->max = 0;
	mgmt->queue = NULL;
	mgmt->queue_depth = 0;
	mgmt->queue_begin = 0;
	mgmt->queue_end = 0;
}

/**
 * Finds the first zero bit in byte from off-th bit.
 * Returns zero on success.
 */
static
int rid_alloc_first_zero(uint32_t *bitmap, size_t w, int off, rid_t *rid)
{
	const rid_t base = w * 8 * sizeof(bitmap[w]);

	for(int i = 0; i < 8; ++i) {
		const uint32_t mask = 1 << (off + i);

		if((bitmap[w] & mask) == 0) {
			bitmap[w] |= mask; // mark bit as used
			*rid = base + off + i;
			return 0;
		}
	}

	abort(); // this should never happen
	return 1;
}

int rid_mgmt_alloc(struct rid_mgmt *mgmt, rid_t *rid)
{
	const size_t words = mgmt->max / 32 + (mgmt->max % 32? 1 : 0);
	uint32_t *bitmap = mgmt->bitmap;

	for(size_t i = 0; i < words; ++i) {
		if(bitmap[i] == 0xFFFFFFFF)
			continue;

		if((bitmap[i] & 0x000000FF) != 0x000000FF)
			return rid_alloc_first_zero(bitmap, i,  0, rid);

		if((bitmap[i] & 0x0000FF00) != 0x0000FF00)
			return rid_alloc_first_zero(bitmap, i,  8, rid);

		if((bitmap[i] & 0x00FF0000) != 0x00FF0000)
			return rid_alloc_first_zero(bitmap, i, 16, rid);

		if((bitmap[i] & 0xFF000000) != 0xFF000000)
			return rid_alloc_first_zero(bitmap, i, 24, rid);

		abort(); // this should never happen
	}

	if(queue_empty(mgmt))
		return 1; // no RID available

	*rid = queue_take(mgmt);
	return 0;
}

void rid_mgmt_release(struct rid_mgmt *mgmt, rid_t rid)
{
	assert(rid < mgmt->max);

	const size_t w = rid / 32;
	const size_t i = rid % 32;

	uint32_t *bitmap = mgmt->bitmap;
	bitmap[w] &= ~(1 << i);
}

void rid_mgmt_release_later(struct rid_mgmt *mgmt, rid_t rid)
{
	if(queue_full(mgmt))
		rid_mgmt_release(mgmt, queue_take(mgmt));

	queue_push_unique(mgmt, rid);
}

#ifdef RID_MGMT_TEST
#include "../cccid_test.h"

static
void test_rid_mgmt_init(void)
{
	struct rid_mgmt mgmt;

	//////////////////////////////////

	memset(&mgmt, 0, sizeof(mgmt));

	if(rid_mgmt_init(&mgmt, 10))
		test_failed("Failed to init rid_mgmt of max 10");

	assert(mgmt.bitmap != NULL);
	assert(*((uint32_t *) mgmt.bitmap) == 0xFFFFFC00);
	assert(mgmt.max == 10);

	rid_mgmt_fini(&mgmt);
	test_ok("Successfully initialized rid_mgmt for max 10");

	//////////////////////////////////

	memset(&mgmt, 0, sizeof(mgmt));

	if(rid_mgmt_init(&mgmt, 65))
		test_failed("Failed to init rid_mgmt of max 65");

	assert(mgmt.bitmap != NULL);
	assert(((uint32_t *) mgmt.bitmap)[2] == 0xFFFFFFFE);
	assert(mgmt.max == 65);

	rid_mgmt_fini(&mgmt);
	test_ok("Successfully initialized rid_mgmt for max 65");

	//////////////////////////////////

	memset(&mgmt, 0, sizeof(mgmt));

	if(rid_mgmt_init(&mgmt, 128))
		test_failed("Failed to init rid_mgmt of max 128");

	assert(mgmt.bitmap != NULL);
	assert(((uint32_t *) mgmt.bitmap)[3] == 0x00000000);
	assert(mgmt.max == 128);

	rid_mgmt_fini(&mgmt);
	test_ok("Successfully initialized rid_mgmt for max 128");
}

static
void test_rid_alloc_first_zero()
{
	rid_t rid;
	uint32_t a[] = {
		0, 0, 0, 0
	};

	for(int i = 0; i < 4; ++i) {
		for(size_t j = 0; j < sizeof(uint32_t) * 8; ++j) {
			int off = (j / 8) * 8;
			if(rid_alloc_first_zero(a, i, off, &rid))
				test_failed("Failed to alloc rid %d "
						"(in 0x%" PRIx32 ")", j, a[i]);

			assert(rid == 8 * sizeof(a[i]) * i + j);
			assert(a[i] & (1 << j));
		}

		test_ok("Allocated %d RIDs", (i + 1) * sizeof(uint32_t) * 8);
	}
}

static
void test_rid_mgmt_alloc(void)
{
	uint32_t bitmap[] = {0, 0, 0, 0};
	struct rid_mgmt mgmt = {
		.bitmap = bitmap,
		.max = sizeof(bitmap) * 8
	};
	rid_t rid;

	for(size_t i = 0; i < mgmt.max; ++i) {
		if(rid_mgmt_alloc(&mgmt, &rid))
			test_failed("Failed to alloc rid: %zu", i);

		assert(rid == i);
	}

	test_ok("Successfully allocated %zu RIDs", mgmt.max);

	if(rid_mgmt_alloc(&mgmt, &rid))
		test_ok("No more room for RIDs, that's right");
	else
		test_failed("Allocation passed even with full RID pool");
}

static
void test_rid_mgmt_release(void)
{
	uint32_t bitmap[] = {~0, ~0, ~0, ~0};
	struct rid_mgmt mgmt = {
		.bitmap = bitmap,
		.max = sizeof(bitmap) * 8
	};

	for(size_t i = 0; i < mgmt.max; ++i) {
		const uint32_t mask = 1 << (i % (sizeof(uint32_t) * 8));
		assert(bitmap[i / (sizeof(uint32_t) * 8)] & mask);

		rid_mgmt_release(&mgmt, (rid_t) i);

		assert((bitmap[i / (sizeof(uint32_t) * 8)] & mask) == 0);
	}

	test_ok("Release of all RIDs was successful");
}

static
void test_random_alloc_release(void)
{
	uint32_t bitmap[] = {~0, 0xFFF0FFFF, 0xFe7b071F, ~0};
	struct rid_mgmt mgmt = {
		.bitmap = bitmap,
		.max = sizeof(bitmap) * 8
	};

	size_t count = 0;
	rid_t rid;

	while(!rid_mgmt_alloc(&mgmt, &rid))
		count++;

	assert(count == 4 + 3 + 4 + 1 + 3);
	test_ok("Successfully allocated the rest");

	for(int i = 0; i < mgmt.max / 4; ++i)
		rid_mgmt_release(&mgmt, i * 3);

	for(int i = 0; i < mgmt.max / 4; ++i) {
		if(rid_mgmt_alloc(&mgmt, &rid))
			test_failed("The alloc failed too early: %d", i);
	}

	test_ok("Successfully allocated the released RIDs");
}

void test_rid_mgmt_less_then32(void)
{
	struct rid_mgmt mgmt;
	if(rid_mgmt_init(&mgmt, 31))
		test_failed("Failed to initialize the RID management for 31 elements");

	rid_t rid;

	for(int i = 0; i < 31; ++i) {
		if(rid_mgmt_alloc(&mgmt, &rid))
			test_failed("Failed to alloc the %dth RID", i);
	}

	test_ok("Successfully allocated all the %d RIDs", 31);

	rid_mgmt_fini(&mgmt);
}

static
void test_simple_release_later(void)
{
	struct rid_mgmt mgmt;
	if(rid_mgmt_init(&mgmt, 128))
		test_failed("Failed to initialize the RID management");

	uint32_t *bitmap = (uint32_t *) mgmt.bitmap;
	rid_t rid;

	///////////////////////////////

	if(rid_mgmt_alloc(&mgmt, &rid))
		test_failed("Failed to alloc the first RID");

	assert(bitmap[0] == 0x01);
	assert(rid == 0);

	rid_mgmt_release(&mgmt, rid);
	assert(bitmap[0] == 0x00);
	assert(mgmt.queue_begin == 0);
	assert(mgmt.queue_end == 0);

	// now do it again, but release later

	if(rid_mgmt_alloc(&mgmt, &rid))
		test_failed("Failed to alloc the first RID again");

	assert(bitmap[0] == 0x01);
	assert(rid == 0);

	rid_mgmt_release_later(&mgmt, rid);
	assert(bitmap[0] == 0x01);
	assert(mgmt.queue[0] == 0);
	assert(mgmt.queue_begin == 1);
	assert(mgmt.queue_end == 0);

	test_ok("Release later of 0 was successful");

	// try alloc another one

	if(rid_mgmt_alloc(&mgmt, &rid))
		test_failed("Failed to alloc the second RID");

	assert(bitmap[0] == 0x03);
	assert(rid == 1);

	rid_mgmt_release_later(&mgmt, rid);
	assert(bitmap[0] == 0x03);
	assert(mgmt.queue[1] == 1);
	assert(mgmt.queue_begin == 2);
	assert(mgmt.queue_end == 0);

	test_ok("Release later of 1 was successful");

	///////////////////////////////

	// avoid duplicates in queue
	
	rid_mgmt_release_later(&mgmt, rid);

	assert(bitmap[0] == 0x03);
	assert(mgmt.queue[1] == 1);
	assert(mgmt.queue_begin == 2);
	assert(mgmt.queue_end == 0);

	test_ok("The duplicate was not inserted again, that's right");

	///////////////////////////////

	rid_mgmt_fini(&mgmt);
}

static
void test_extreme_release_later(void)
{
	struct rid_mgmt mgmt;
	if(rid_mgmt_init(&mgmt, 16))
		test_failed("Failed to initialize the RID management");

	uint32_t *bitmap = (uint32_t *) mgmt.bitmap;
	rid_t rid;

	///////////////////////////////

	for(int i = 0; i < 16; ++i) {
		if(rid_mgmt_alloc(&mgmt, &rid))
			test_failed("Failed to alloc the %dth RID", i);
	}

	test_ok("Allocated 16 RIDs to test release_later");

	// release 4 immediately and check

	assert(bitmap[0] = 0xFFFFFFFF);
	rid_mgmt_release(&mgmt, 4);
	assert(bitmap[0] = 0xFFFFFFEF);

	// alloc again

	if(rid_mgmt_alloc(&mgmt, &rid))
		test_failed("Failed to alloc back the RID %" PRIu16, rid);

	test_ok("Unnecessary checking of release/alloc successful");

	// release 4 later

	assert(bitmap[0] == 0xFFFFFFFF);
	rid_mgmt_release_later(&mgmt, 4);
	assert(bitmap[0] == 0xFFFFFFFF);

	assert(mgmt.queue[0] == 4);
	assert(mgmt.queue_begin == 1);
	assert(mgmt.queue_end == 0);

	assert(!queue_full(&mgmt));

	// release 11 later

	assert(bitmap[0] == 0xFFFFFFFF);
	rid_mgmt_release_later(&mgmt, 11);
	assert(bitmap[0] == 0xFFFFFFFF);

	assert(mgmt.queue[1] == 11);
	assert(mgmt.queue_begin == 2);
	assert(mgmt.queue_end == 0);

	assert(!queue_full(&mgmt));

	// release 7 later

	assert(bitmap[0] == 0xFFFFFFFF);
	rid_mgmt_release_later(&mgmt, 7);
	assert(bitmap[0] == 0xFFFFFFFF);

	assert(mgmt.queue[2] == 7);
	assert(mgmt.queue_begin == 3);
	assert(mgmt.queue_end == 0);

	assert(!queue_full(&mgmt));

	// release 1 later (last, just 4 rooms)

	assert(bitmap[0] == 0xFFFFFFFF);
	rid_mgmt_release_later(&mgmt, 1);
	assert(bitmap[0] == 0xFFFFFFFF);

	assert(mgmt.queue[3] == 1);
	assert(mgmt.queue_begin == 4);
	assert(mgmt.queue_end == 0);

	assert(queue_full(&mgmt));

	// release 2 later and return 4 to the pool
	
	assert(bitmap[0] == 0xFFFFFFFF);
	rid_mgmt_release_later(&mgmt, 2);
	assert(bitmap[0] == 0xFFFFFFEF);

	assert(mgmt.queue[4] == 2);
	assert(mgmt.queue_begin == 0);
	assert(mgmt.queue_end == 1);

	assert(queue_full(&mgmt));

	test_ok("Successfully performed release_later for 5 RIDs and kicked out the oldest");

	///////////////////////////////
	
	// alloc from pool

	assert(bitmap[0] == 0xFFFFFFEF);

	if(rid_mgmt_alloc(&mgmt, &rid))
		test_failed("Failed to alloc the only one available RID %d", 4);

	assert(bitmap[0] == 0xFFFFFFFF);
	assert(rid == 4);

	// alloc from queue

	assert(bitmap[0] == 0xFFFFFFFF);
	assert(queue_full(&mgmt));

	if(rid_mgmt_alloc(&mgmt, &rid))
		test_failed("Failed to alloc from queue");

	assert(bitmap[0] == 0xFFFFFFFF);
	assert(rid == 11);

	assert(mgmt.queue[1] == rid);
	assert(mgmt.queue_begin == 0);
	assert(mgmt.queue_end == 2);

	assert(!queue_full(&mgmt));

	test_ok("Successfully allocated from queue");

	///////////////////////////////

	rid_mgmt_fini(&mgmt);
}

int main(void)
{
	test_rid_mgmt_init();
	test_rid_alloc_first_zero();
	test_rid_mgmt_alloc();
	test_rid_mgmt_release();
	test_random_alloc_release();
	test_rid_mgmt_less_then32();
	test_simple_release_later();
	test_extreme_release_later();
	return 0;
}

#endif
