/*!
 * \file delayctl.c
 * \brief Tool for controlling delay units
 * \author Lukas Kekely <xkekel00@stud.fit.vutbr.cz>
 * \date 2012
 *
 * Copyright (C) 2012 Brno University of Technology
 *
 * LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * $Id: delayctl.c 2596 2012-10-11 15:48:10Z xkekel00 $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <ctype.h>
#include "hwio.h"
#include "hwio_tool.h"

#define TOOL_OPTS "DSt:i:"

#define DELAY_MAX_LEN 23
typedef struct {
    uint32_t s;
    uint32_t ns;
} delay_t;

struct hwio_tool_custom {
    delay_t delay;
    int interface;
};

#define DTREE_PATH "/proc/device-tree"
#define INTERFACES    2    // number of interfaces

struct hwio_comp_spec tool_compat[]={
    {"combov2", "Delay_buffer", HWIO_VER2(1, 0)},
    HWIO_COMPAT_END
};

void hwio_tool_help(const char *toolname, const struct hwio_tool_params *params) {
    printf("Usage: %s [options | commands]\n", toolname);
    printf("Options:\n");
    printf("  -h     prints this help message\n");
    printf("  -v     verbosity level\n");
    printf("  -q     no verbose (negative verbosity level)\n");
    printf("  -t del defines value of delay for \"D\" command (default: none)\n");
    printf("  -i ifc defines interface for \"D\" command to be used (default: all interfaces) \n");
    printf("  -x pth path to design.xml file\n");
    printf("  -d pth path to device file\n");
    printf("  -A     perform action on all delay buffers in design\n");
    printf("Commands:\n");
    printf("  -D     set delay on interfaces (specified by \"t\" and \"i\")\n");
    printf("  -I     informations about all delay buffers in design\n");
    printf("  -L     list all delay buffers in design\n");
    printf("  -S     show configured delay on all interfaces\n");
    printf("Where:\n");
    printf("  del    delay time with 1ns accuracy and <0,2^25s) span\n");
    printf("         Format: [[[dd:]hh:]mm:]ss[.ss]\n");
    printf("         Special: 'none'  - minimum delay (0)\n");
    printf("                 'max'   - maximum delay (388:08:40:31.999999999)\n");
    printf("  ifc    interface number starting from 0\n");
}

void show_delay(const delay_t *dl, char *buff) {
    int d,h,m,s;
    if(dl->s==33525631&&dl->ns==999999999) {
        sprintf(buff,"max");
        return;
    }
    s=dl->s;
    d=s/86400;  //days
    s=s%86400;
    h=s/3600;   //hours
    s=s%3600;
    m=s/60;     //minutes
    s=s%60;
    if(d)
        sprintf(buff,"%d:%02d:%02d:%02d",d,h,m,s);
    else if(h)
        sprintf(buff,"%d:%02d:%02d",h,m,s);
    else if(m)
        sprintf(buff,"%d:%02d",m,s);
    else if(s||dl->ns)
        sprintf(buff,"%d",s);
    else
        sprintf(buff,"none");
    buff+=strlen(buff);
    if(dl->ns)
        sprintf(buff,".%09d",dl->ns);
}

int get_unsigned(const char *s, uint32_t max, uint32_t *res) {
    char *e;
    *res=strtoul(s,&e,0);
    return ((*e)=='\0'&&(*res)<max);
}

int get_ifc(const char *s) {
    uint32_t x;
    if(get_unsigned(s,INTERFACES,&x))
        return x;
    return -1;
}

int split_char(char *str, char ch) {
    int c=0;
    while((*str)!='\0') {
        if((*str)==ch) {
            *str='\0';
            c++;
        }
        str++;
    }
    return c;
}

int get_delay(delay_t *dl, const char *cstr) {
    int size=strlen(cstr)+1,parts;
    uint32_t d=0,h=0,m=0,s=0,ns=0,max;
    char buff[size];
    char *str=buff;
    memcpy(str,cstr,size);
    for(;*str!='\0'; str++)
        *str=tolower(*str);
    str=buff;
    if(strcmp(str,"none")==0) { // special value: "none"
        dl->s=0; dl->ns=0;
    } else if(strcmp(str,"max")==0) { // special value: "max"
        dl->s=33525631; dl->ns=999999999;
    } else { // basic time format
        parts=split_char(str,':'); // format: [[[d:]h:]m:]s[.s]
        if(parts>3)
            return 0;
        if(parts==3) {  // days
            if(!get_unsigned(str,389,&d))
                return 0;
            max=24;
            str+=strlen(str)+1;
        } else
            max=9321;
        if(parts>=2) {  // hours
            if(!get_unsigned(str,max,&h))
                return 0;
            max=60;
            str+=strlen(str)+1;
        } else
            max=558761;
        if(parts>=1) {  // minutes
            if(!get_unsigned(str,max,&m))
                return 0;
            max=60;
            str+=strlen(str)+1;
        } else
            max=33525632;
        parts=split_char(str,'.'); // format: s[.s]
        if(parts>1)
            return 0;
        if(!get_unsigned(str,max,&s)) // seconds
            return 0;
        dl->s=(((d*24)+h)*60+m)*60+s;
        if(dl->s>=33525632)
            return 0;
        if(parts==1) {
            str+=strlen(str)+1;
            if(!get_unsigned(str,1000000000,&ns)) // nanoseconds
                return 0;
            for(int i=strlen(str); i<9; i++)
                ns*=10;
        }
        dl->ns=ns;
    }
    return 1;
}

/// set delay for specified interface
int delay_set(struct hwio_comp *c, hwio_tool_comp_err_t err, const struct hwio_tool_params *params) {
    int sel=params->custom->interface;
    delay_t *d=&(params->custom->delay);
    for(int i=0; i<INTERFACES; i++) {
        if(i==sel||sel==-1) {
            if(hwio_comp_write(c,(i<<3)+4,d->s) !=HWIO_EOK ||
               hwio_comp_write(c,i<<3    ,d->ns)!=HWIO_EOK) {
                fprintf(stderr,"Unable to write configuration to delay buffer!\n");
                return EXIT_FAILURE;
            }
        }
    }
    return EXIT_SUCCESS;
}

/// read delays of all interfaces
int config_show(struct hwio_comp *c, hwio_tool_comp_err_t err, const struct hwio_tool_params *params) {
    delay_t d;
    char buff[DELAY_MAX_LEN];
    printf(" *** DELAY UNITS CONFIGURATION *** \n");
    for(int i=0; i<INTERFACES; i++) {
        if(hwio_comp_read(c,i<<3    ,&(d.ns))!=HWIO_EOK ||
           hwio_comp_read(c,(i<<3)+4,&(d.s)) !=HWIO_EOK) {
            fprintf(stderr,"Unable to read configuration from delay buffer!\n");
            return EXIT_FAILURE;
        }
        show_delay(&d,buff);
        printf("Interface #%d: %s\n",i,buff);
    }
    return EXIT_SUCCESS;
}

int tool_custom_parse(int opt, const char *optarg, struct hwio_tool_params *params) {
    switch(opt) {
        case 'D': case 'S':
            if(hwio_tool_params_has_cmd(params))
                return hwio_tool_error_another_cmd(opt,params);
            params->cmd.code = opt;
            params->cmd.arg = NULL;
            return HWIO_TOOL_CUSTOM_OK;
        case 't':
            if(get_delay(&(params->custom->delay),optarg))
                return HWIO_TOOL_CUSTOM_OK;
            else
                return hwio_tool_error(HWIO_TOOL_EARGS,"Wrong delay time format");
        case 'i':
            params->custom->interface=get_ifc(optarg);
            if(params->custom->interface>=0)
                return HWIO_TOOL_CUSTOM_OK;
            else
                return hwio_tool_error(HWIO_TOOL_EARGS,"Wrong interface number format");
        default:
            return HWIO_TOOL_CUSTOM_UNKNOWN;
    }
}

/// MAIN //////////////////////////////////////////////////////
int main(int argc, char *argv[]) {
    struct hwio_tool_custom tool_params={{0,0},-1};
    struct hwio_tool_params params = hwio_tool_params_new(&tool_params,argc,argv);
    switch(hwio_tool_params_parse(&params,&tool_custom_parse,TOOL_OPTS)) {
        case HWIO_TOOL_PARAMS_FAILED:
            return HWIO_TOOL_EARGS;
        case HWIO_TOOL_PARAMS_EXIT:
            return 0;
        default:
            break;
    }
    hwio_tool_verbose(&params, 1, "Arguments parsed successfully");
    hwio_tool_verbose(&params, 1, "Command: -%c", params.cmd.code);
    hwio_tool_verbose(&params, 2, "Argument: %s", params.cmd.arg);
    switch(params.cmd.code) {
        case 'I':
            return hwio_tool_exec_generic(&params,tool_compat,&hwio_tool_comp_info,&hwio_tool_error_handler);
        case 'L':
            return hwio_tool_exec_generic(&params,tool_compat,&hwio_tool_comp_name,&hwio_tool_error_handler);
        case 'D':
            return hwio_tool_exec_generic(&params,tool_compat,&delay_set,&hwio_tool_error_handler);
        case 'S':
            return hwio_tool_exec_generic(&params,tool_compat,&config_show,&hwio_tool_error_handler);
        case '\0':
            return hwio_tool_error(HWIO_TOOL_EARGS,"No command was specified");
        default:
            return hwio_tool_error(HWIO_TOOL_EARGS,"Command -%c is not implemented",params.cmd.code);
    }
}

