/**
 * @file hwio_cv2_query.c
 * @brief Query functions for libcombo implementation
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_cv2.h"
#include "hwio_private.h"

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <combo.h>

const char VENDOR[] = "combov2";

/**
 * Extract data from a `cs_component_t' structure and store them
 * in an `hwio_comp_t' structure.
 */
static
int hwio_cs_to_comp(const struct hwio_dev *dev, cs_component_t *src,
	const char *src_name, cs_cidx_t src_index, struct hwio_comp *dst)
{
	int retcode = HWIO_EOK;

	dst->plat_comp = src;
	dst->plat_free = &hwio_plat_comp_free;
	dst->dev = dev;

	dst->vendor = strdup(VENDOR);
	if (dst->vendor == NULL)
		return HWIO_ENOMEM;

	dst->type = strdup(src_name);
	if (dst->type == NULL) {
		retcode = HWIO_ENOMEM;
		goto cleanup_vendor;
	}

	int index_len = (src_index == 0) ? 1 : floor(log10(src_index)) + 1;
	dst->name = malloc(index_len + 1);
	if (dst->name == NULL) {
		retcode = HWIO_ENOMEM;
		goto cleanup_type;
	}

	snprintf(dst->name, index_len + 1, "%d", src_index);
	cs_cver_t cver = cs_component_version(src);
	dst->version = (struct hwio_version) HWIO_VER2(CS_CVER_MAJOR(cver),
		CS_CVER_MINOR(cver));
	dst->next = NULL;

	return HWIO_EOK;
cleanup_type:
	free(dst->type);
cleanup_vendor:
	free(dst->vendor);
	return retcode;
}

int hwio_comp_find_spec(struct hwio_dev *dev,
	const struct hwio_comp_spec *spec, const char *name,
	struct hwio_comp **head)
{
	cs_device_t *cdev = (cs_device_t *) dev->plat_dev;
	if (cdev == NULL)
		return HWIO_EINVAL;

	int retcode = HWIO_EOK;
	cs_component_t *cs_comp;
	cs_cidx_t cs_index = 0;

	while (cs_component_find(cdev, &cs_comp, NULL, spec->type, cs_index) == 0) {
		struct hwio_comp hwio_comp;

		retcode = hwio_cs_to_comp(dev, cs_comp, spec->type, cs_index, &hwio_comp);
		if (retcode != HWIO_EOK)
			goto cleanup_list;

		struct hwio_comp_spec hwio_spec;
		retcode = hwio_comp_to_spec(&hwio_comp, &hwio_spec);
		if (retcode != HWIO_EOK)
			goto cleanup_list;

		if (hwio_comp_name_and_spec_match(&hwio_spec, hwio_comp.name, spec, name)) {
			struct hwio_comp *comp = hwio_comp_alloc();

			if (comp != NULL)
				memcpy(comp, &hwio_comp, sizeof(hwio_comp));

			if (comp == NULL || hwio_comp_append_if_unique(head, comp) != HWIO_EOK) {
				retcode = HWIO_ENOMEM;
				goto cleanup_list;
			}
		}

		cs_index++;
	}

	return HWIO_EOK;
cleanup_list:
	hwio_comp_list_free(*head);
	return retcode;
}
