/**
 * @file hwio_debug.h
 * @brief Debug implementation using memory and file
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#ifndef _HWIO_DEBUG_H
#define _HWIO_DEBUG_H

#include "hwio_error.h"
#include "hwio_comp.h"
#include "hwio_private.h"

#include <stddef.h>
#include <stdint.h>

/**
 * Platform vendor name.
 */
extern const char VENDOR[];

/**
 * Supported chip in XML file.
 */
#define XML_CHIP "xc5vlx155t"

/**
 * XPath expression to parse component info from XML file.
 */
extern const char XML_XPATH[];

/**
 * Platform-dependent device structure.
 */
struct hwio_plat_file {
	const char *filename;
	uint32_t *memory;
	size_t memory_size;
	struct hwio_comp *comps;
};

/**
 * Platform specific part of the hwio_comp (plat_comp).
 */
struct hwio_debug_comp {
	uint32_t base;
	uint32_t range;
};

/**
 * Allocate hwio_comp and hwio_debug_comp structures at once
 * and link them.
 */
struct hwio_comp *hwio_comp_alloc(void);

static inline
struct hwio_debug_comp *hwio_comp_plat(struct hwio_comp *c)
{
	return (struct hwio_debug_comp *) c->plat_comp;
}

static inline
uint32_t hwio_debug_base(struct hwio_comp *c)
{
	return hwio_comp_plat(c)->base;
}

static inline
uint32_t hwio_debug_range(struct hwio_comp *c)
{
	return hwio_comp_plat(c)->range;
}

/**
 * Free libxml2 text.
 */
void hwio_xml_free(char *text);

/**
 * Load information about components of a design from an XML file.
 * @param src XML file with a description of the design.
 * @param xpath_expr XPath expression to find component nodes.
 * @param head The head of resulting list with component information.
 * @return HWIO_EOK on success, HWIO_EINFO on error during gathering
 * of information error, or the return code of the function `hwio_xml_list'.
 * @see hwio_xml_list
 */
int hwio_xml_load(struct hwio_dev *dev, const char *src,
	const char *xpath_expr,	struct hwio_comp **head);

/**
 * Deallocate platform-dependent component data.
 */
void hwio_plat_comp_free(struct hwio_comp *comp);

#endif /* _HWIO_DEBUG_H */
