/**
 * @file hwio_debug_dev.c
 * @brief Device functions for debug implementation
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_debug.h"
#include "hwio_private.h"

#include <stdlib.h> 
#include <string.h>
#include <stdio.h>

const char *hwio_default_dev_phys(void)
{
	return "fake-component-data";
}

const char *hwio_default_dev_src(void)
{
	return "Design.xml";
}

/**
 * Find the last address in the address space of the device,
 * i.e. the last address of the address space of the last component.
 */
static
size_t hwio_find_last(struct hwio_comp *head)
{
	size_t last = 0;

	while (head != NULL) {
		size_t comp_last = hwio_debug_base(head) + hwio_debug_range(head);

		if (comp_last > last)
			last = comp_last;

		head = hwio_comp_next(head);
	}

	return last;
}

/**
 * Allocate an `hwio_plat_file' structure and corresponding memory
 * to emulate the address space of the device.
 * @param dev Device to initialize.
 * @param size Size of the memory for the address space of the device.
 * @return HWIO_EOK on success or HWIO_ENOMEM on allocation error.
 */
static
int hwio_plat_file_init(struct hwio_plat_file **dev, size_t size)
{
	*dev = malloc(sizeof(**dev));
	if (dev == NULL)
		return HWIO_ENOMEM;

	(*dev)->memory_size = size;
	(*dev)->memory = malloc(size);

	if ((*dev)->memory == NULL) {
		free(*dev);
		*dev = NULL;
		return HWIO_ENOMEM;
	}

	return HWIO_EOK;
}

/**
 * Load a binary file into the memory for the address space of a device.
 * @return HWIO_EOK on success, HWIO_EATT if the binary file cannot be opened,
 *   or HWIO_EINAL if the binary file is smaller than the address space.
 */
static
int hwio_plat_file_load(struct hwio_plat_file *dev, const char *filename)
{
	dev->filename = filename;
	FILE *input = fopen(filename, "r");
	if (input == NULL)
		return HWIO_EATT;

	fseek(input, 0, SEEK_END);
	(void) fseek(input, 0, SEEK_SET);
	size_t read = fread(dev->memory, 1, dev->memory_size, input);
	int result = HWIO_EOK;

	if (read < dev->memory_size)
		result = HWIO_EINVAL;

	fclose(input);
	return result;
}

/**
 * 1. Load information about components from `src'.
 * 2. Allocate memory representing address space.
 * 3. Load data from `phys' to the memory.
 */
int hwio_init(struct hwio_dev *dev, const char *phys, const char *src)
{
	if(phys == NULL || src == NULL)
		return HWIO_EINVAL;

	struct hwio_comp *head = NULL;
	int result = hwio_xml_load(dev, src, XML_XPATH, &head);
	if (result != HWIO_EOK)
		return result;

	struct hwio_plat_file *plat_file;
	result = hwio_plat_file_init(&plat_file, hwio_find_last(head));
	if (result != HWIO_EOK)
		goto cleanup_comp_list;
	dev->plat_dev = plat_file;

	plat_file->comps = head;

	result = hwio_plat_file_load(plat_file, phys);
	if (result != HWIO_EOK)
		goto cleanup_plat_file;

	return HWIO_EOK;
cleanup_plat_file:
	free(plat_file->memory);
	free(plat_file);
cleanup_comp_list:
	hwio_comp_list_free(head);
	return result;
}

/**
 * 1. Store data from the memory to the output file.
 * 2. Free the memory.
 * 3. Free information about components.
 */
void hwio_deinit(struct hwio_dev *dev)
{
	struct hwio_plat_file *plat_file = dev->plat_dev;
	FILE *output = fopen(plat_file->filename, "w");

	if (output != NULL) {
		fwrite(plat_file->memory, 1, plat_file->memory_size, output);
		fclose(output);
	}

	free(plat_file->memory);
	hwio_comp_list_free(plat_file->comps);
	free(plat_file);
	dev->plat_dev = NULL;
}
