/**
 * @file hwio_debug_query.c
 * @brief Query functions for debug implementation
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_debug.h"
#include "hwio_private.h"

#include <stdlib.h> 
#include <string.h>
#include <stdio.h>

/**
 * Append new allocated item to the list only if it is not already contained.
 * The data is deep-copied from the existing component.
 * @param head The list (its first item) to append to.
 *	 If NULL, the new list (its first item) is allocated and returned.
 * @param src The filled component.
 * @return Returns the new allocated and filled component or NULL
 *	 on allocation error.
 */
static
struct hwio_comp *hwio_comp_list_append(struct hwio_comp **head,
	const struct hwio_comp *src)
{
	struct hwio_comp *dst = hwio_comp_alloc();
	if (dst == NULL)
		return NULL;

	memcpy(dst, src, sizeof(*src));

	dst->vendor = strdup(src->vendor);
	if (dst->vendor == NULL)
		goto cleanup_dst;

	dst->type = strdup(src->type);
	if (dst->type == NULL)
		goto cleanup_dst_vendor;

	dst->name = strdup(src->name);
	if (dst->name == NULL)
		goto cleanup_dst_type;

	if (hwio_comp_append_if_unique(head, dst) != HWIO_EOK)
		goto cleanup_dst_type;

	return dst;

cleanup_dst_type:
	free(dst->type);
cleanup_dst_vendor:
	free(dst->vendor);
cleanup_dst:
	free(dst);
	return NULL;
}

int hwio_comp_find_spec(struct hwio_dev *dev,
	const struct hwio_comp_spec *spec, const char *name,
	struct hwio_comp **head)
{
	struct hwio_plat_file *plat_file = dev->plat_dev;
	if (plat_file == NULL)
		return HWIO_EINVAL;
	struct hwio_comp *hwio_comp = plat_file->comps;

	while (hwio_comp != NULL) {
		struct hwio_comp_spec hwio_spec;

		int retcode = hwio_comp_to_spec(hwio_comp, &hwio_spec);
		if (retcode != HWIO_EOK)
			return retcode;

		if (hwio_comp_name_and_spec_match(&hwio_spec, hwio_comp->name, spec, name)) {
			if (hwio_comp_list_append(head, hwio_comp) == NULL)
				return HWIO_ENOMEM;
		}

		hwio_comp = hwio_comp_next(hwio_comp);
	}

	return HWIO_EOK;
}
