/**
 * @file hwio_debug_xml.c
 * @brief XML functions for debug implementation
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_debug.h"
#include "hwio_private.h"

#include <stdlib.h> 
#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#if !defined(LIBXML_XPATH_ENABLED) || !defined(LIBXML_SAX1_ENABLED)
#error "libxml2 needs to be compiled with XPath and SAX1 support"
#endif

const char VENDOR[] = "debug";

const char XML_XPATH[] = "/combo6design/firmware[@chip='" XML_CHIP "']/component";

/**
 * Extract data from an XML component node and store them
 * in an `hwio_comp' structure.
 * @param src XML component node.
 * @return Returns HWIO_EOK on succes or HWIO_EINVAL on missing
 *   XML attribute (one of `cversion', `name', `index', `base' or `size').
 */
static
int hwio_xml_extract(struct hwio_dev *dev, xmlNodePtr src,
	struct hwio_comp *dst)
{
	int retcode = HWIO_EOK;
	xmlChar *version = xmlGetProp(src, BAD_CAST "cversion");
	xmlChar *type = xmlGetProp(src, BAD_CAST "name");
	xmlChar *name = xmlGetProp(src, BAD_CAST "index");
	xmlChar *base = xmlGetProp(src, BAD_CAST "base");
	xmlChar *range = xmlGetProp(src, BAD_CAST "size");
	if (version == NULL || type == NULL || name == NULL ||
		base == NULL || range == NULL)
		return HWIO_EINVAL;

	dst->plat_free = &hwio_plat_comp_free;
	dst->dev = dev;

	dst->vendor = strdup(VENDOR);
	if (dst->vendor == NULL) {
		retcode = HWIO_ENOMEM;
		goto cleanup_all_props;
	}

	dst->type = (char *) type;
	dst->name = (char *) name;

	sscanf((char *) version, "%d.%d", &dst->version.major,
		&dst->version.minor);
	dst->version.subminor = 0;

	hwio_comp_plat(dst)->base = (uint32_t) strtol((char *) base, NULL, 0);
	hwio_comp_plat(dst)->range = (uint32_t) strtol((char *) range, NULL, 0);
	dst->next = NULL;

	goto cleanup_normal;

cleanup_all_props:
	xmlFree(name);
	xmlFree(type);
cleanup_normal:
	xmlFree(range);
	xmlFree(base);
	xmlFree(version);
	return retcode;
}

/**
 * Create a list of `hwio_comp' structures
 * from a set of XML component nodes.
 * @param nodes Set of XML component nodes.
 * @param head The head of resulting list.
 * @return Returns HWIO_EOK on success, HWIO_ENOMEM on allocation error
 *   or HWIO_EINVAL on missing XML attribute.
 */
static
int hwio_xml_list(struct hwio_dev *dev, xmlNodeSetPtr nodes,
	struct hwio_comp **head)
{
	*head = NULL;
	int retcode = HWIO_EOK;
	int n = (nodes != NULL) ? nodes->nodeNr : 0;
	struct hwio_comp *last = NULL;
	int i;

	for (i = 0; i < n; i++) {
		xmlNodePtr node = nodes->nodeTab[i];

		struct hwio_comp *comp = hwio_comp_alloc();
		if (comp == NULL) {
			retcode = HWIO_ENOMEM;
			goto cleanup_list;
		}

		retcode = hwio_xml_extract(dev, node, comp);
		if (retcode != HWIO_EOK) {
			free(comp);
			goto cleanup_list;
		}

		if (last == NULL)
			*head = comp;
		else
			last->next = comp;

		last = comp;
	}
	return HWIO_EOK;
cleanup_list:
	hwio_comp_list_free(*head);
	*head = NULL;
	return retcode;
}

void hwio_xml_free(char *text)
{
	xmlFree(BAD_CAST text);
}

int hwio_xml_load(struct hwio_dev *dev, const char *src,
	const char *xpath_expr,	struct hwio_comp **head)
{
	xmlInitParser();
	LIBXML_TEST_VERSION

	int result;

	xmlDocPtr xml_doc = xmlParseFile(src);
	if (xml_doc == NULL) {
		result = HWIO_EINFO;
		goto cleanup_parser;
	}

	xmlXPathContextPtr xpath_context = xmlXPathNewContext(xml_doc);
	if (xpath_context == NULL) {
		result = HWIO_EINFO;
		goto cleanup_doc;
	}

	xmlXPathObjectPtr xresult = xmlXPathEvalExpression(BAD_CAST xpath_expr,
		xpath_context);
	if (xresult == NULL) {
		result = HWIO_EINFO;
		goto cleanup_context;
	}

	result = hwio_xml_list(dev, xresult->nodesetval, head);	

	xmlXPathFreeObject(xresult);
cleanup_context:
	xmlXPathFreeContext(xpath_context);
cleanup_doc:
	xmlFreeDoc(xml_doc);
cleanup_parser:
	xmlCleanupParser();

	return result;
}
