/**
 * @file hwio_linux.h
 * @brief Non-public header for the linux implementation
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#ifndef _HWIO_LINUX
#define _HWIO_LINUX

#ifndef _BSD_SOURCE
#define _BSD_SOURCE
#endif

#include "hwio_dev.h"
#include "hwio_comp.h"
#include "hwio_private.h"

#include <stdint.h>
#include <unistd.h>

/**
 * Holds information that is necessary to open, query and close
 * the dtree interface and to mmap address space of a component.
 */
struct hwio_linux_dev {
	char *dev_file;
	char *dtree_root;
};

/**
 * Simple getter to convert the plat_dev pointer to the hwio_linux_dev.
 */
static inline
struct hwio_linux_dev *as_linux_dev(const struct hwio_dev *dev)
{
	return (struct hwio_linux_dev *) dev->plat_dev;
}

/**
 * Platform dependent hwio_comp structure.
 * It is stored in hwio_comp->plat_comp.
 */
struct hwio_linux_comp {
	uint32_t base;
	uint32_t range;

	int mmap_fd;
	void *mmap_m;
	uint32_t mmap_siz;
};

/**
 * Simple getter to convert the plat_comp pointer to the hwio_linux_comp.
 */
static inline
struct hwio_linux_comp *as_linux_comp(const struct hwio_comp *comp)
{
	return (struct hwio_linux_comp *) comp->plat_comp;
}

/**
 * Allocates a new hwio_comp instance.
 */
struct hwio_comp *hwio_comp_alloc(const char *name,
		uint32_t base, uint32_t range);

/**
 * Computes the alignment of the base for mmap call.
 */
static inline
uint32_t get_alignment(uint32_t base)
{
	return base % getpagesize();
}

/**
 * Platform specific free for hwio_comp (hwio_comp->plat_free).
 * There is not other usage of this function.
 */
void hwio_linux_comp_free(struct hwio_comp *c);

/**
 * Maps the address space into memory of the process.
 * If the space is already mapped, it does nothing.
 * Returns 0 when successful.
 */
int map_space_if_needed(struct hwio_linux_dev *d,
		struct hwio_linux_comp *c);

/**
 * Unmaps the space.
 */
void unmap_space_if_mapped(struct hwio_linux_comp *c);

#endif
