/**
 * @file hwio_linux_dev.c
 * @brief Implementation of hwio_dev interface
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_linux.h"
#include "hwio.h"

#include <stdlib.h>
#include <string.h>

static
int hwio_linux_init(struct hwio_dev *dev, const char *dev_file, const char *dtree_root)
{
	const size_t dtree_root_len = strlen(dtree_root) + 1;
	const size_t dev_file_len   = strlen(dev_file) + 1;
	const size_t msiz = sizeof(struct hwio_linux_dev) + dtree_root_len + dev_file_len;

	// allocate all at once
	void *m = malloc(msiz);
	if(m == NULL)
		return 1;

	// Memory layout:
	//  ------------------------------------------------
	// | hwio_linux_dev | dev_file | Z | dtree_root | Z |
	//  ------------------------------------------------
	struct hwio_linux_dev *impl = (struct hwio_linux_dev *) m;
	impl->dev_file   = (char *) (impl + 1);
	impl->dtree_root = impl->dev_file + dev_file_len;

	// copy the paths, the '\0' is included
	memcpy(impl->dev_file, dev_file, dev_file_len);
	memcpy(impl->dtree_root, dtree_root, dtree_root_len);

	dev->plat_dev = impl;
	return 0;
}

static
void hwio_linux_deinit(struct hwio_dev *dev)
{
	if(dev == NULL)
		return;

	if(dev->plat_dev == NULL)
		return;

	free(dev->plat_dev);
	dev->plat_dev = NULL;
}

const char *hwio_default_dev_phys(void)
{
	return "/dev/mem";
}

const char *hwio_default_dev_src(void)
{
	return "/proc/device-tree";
}

int hwio_init(struct hwio_dev *dev, const char *dev_file, const char *dtree_root)
{
	if(dev_file == NULL || dtree_root == NULL)
		return HWIO_EINVAL;

	if(hwio_linux_init(dev, dev_file, dtree_root))
		return HWIO_ENOMEM;

	return HWIO_EOK;
}


void hwio_deinit(struct hwio_dev *dev)
{
	hwio_linux_deinit(dev);
}

///////////////////////////////

#ifdef TEST_HWIO_DEV
#include "hwio_test.h"

int main(int argc, char **argv)
{
	struct hwio_dev dev;
	int ret = hwio_init(&dev, hwio_default_dev_phys(), hwio_default_dev_src());

	subtest_start("hwio_init");
	if(ret != HWIO_EOK) {
		perror("hwio_init");
		test_fail("Failed to hwio_init", ret);
		return 1;
	}
	subtest_end("hwio_init");

	hwio_deinit(&dev);
	return 0;
}
#endif
