/**
 * @file hwio_linux_query.c
 * @brief Implementation of hwio_query interface
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#define _GNU_SOURCE

#include "hwio_private.h"

#include "hwio_comp.h"
#include "hwio_comp_spec.h"
#include "hwio_error.h"
#include "hwio_version.h"
#include "dtree.h"

#include "hwio_linux.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <regex.h>
#include <ctype.h>

static
int do_dtree_open(struct hwio_dev *dev)
{
	const char *rootd = as_linux_dev(dev)->dtree_root;

	if(dtree_open(rootd))
		return -1;

	return 0;
}

static
void do_dtree_close(struct hwio_dev *dev)
{
	dtree_close();
}

/**
 * Parses number in the string s from beg (included) until end (excluded).
 * The result is set to the num.
 * The function assumes that the string range contains a valid number.
 */
static
void parse_num(int *num, const char *s, int beg, int end)
{
	assert(beg >= 0);
	assert(end >= 0);

	int result = 0;
	int factor = 1;

	for(int i = end - 1; i >= beg; --i) {
		assert(isdigit(s[i]));

		result += factor * (s[i] - '0');
		factor *= 10;
	}

	*num = result;
}

/**
 * Parses an alpha character or a number in the string s from beg (included)
 * until end (excluded). If beg or end is negative then HWIO_VERSION_NA will
 * be the result (so the parsing doesn't occur). This allows to make such a
 * value to be optional.
 *
 * Otherwise (exactly) 1 alpha character is used (the range length equals to 1)
 * or a number is parsed.
 */
static
void parse_optional_alpha_or_num(int *num, const char *s, int beg, int end)
{
	if(beg < 0 || end < 0) {
		*num = HWIO_VERSION_NA;
		return;
	}

	if(isalpha(s[beg])) {
		assert(end - beg == 1);
		*num = s[beg];
		return;
	}

	parse_num(num, s, beg, end);
}

/**
 * Parses a version string in the form a.b.c, where a and b are decimal numbers
 * and c is either a number or an alpha character.
 */
static
void parse_version(struct hwio_version *version, const char *vstr)
{
	regex_t regex;

#define X_MAJ "\\([0-9][0-9]*\\)"
#define X_MIN X_MAJ
#define X_SUB "\\([a-z]\\|[0-9][0-9]*\\)"
#define X_DOT "\\."
#define X_HAS "\\{0,1\\}" // this is '?' in extended syntax

	if(regcomp(&regex, "^" X_MAJ X_DOT X_MIN "\\(" X_DOT X_SUB "\\)" X_HAS, 0))
		abort();

	regmatch_t m[5];
	int ret = regexec(&regex, vstr, 5, m, 0);
	switch(ret) {
	case 0:
		parse_num(&version->major, vstr, m[1].rm_so, m[1].rm_eo);
		parse_num(&version->minor, vstr, m[2].rm_so, m[2].rm_eo);
		parse_optional_alpha_or_num(&version->subminor, vstr,
				m[4].rm_so, m[4].rm_eo);
		break;

	case REG_NOMATCH:
		hwio_version_set(version, HWIO_VERSION_NA, HWIO_VERSION_NA, HWIO_VERSION_NA);
		break;

	default:
		abort();
	}

	regfree(&regex);
}

/**
 * Treat all the compatible string to be a type, so just copy
 * the whole string.
 * Allocates the extracted values at once, to free, just do free(c->vendor).
 */
static
int treat_it_as_type(struct hwio_comp_spec *c, const char *type)
{
	const size_t vendor_len = strlen(HWIO_VENDOR_UNKNOWN);
	const size_t type_len = strlen(type);

	c->vendor = malloc(vendor_len + 1 + type_len + 1);
	if(c->vendor == NULL)
		return 1;

	memcpy((void *) c->vendor, HWIO_VENDOR_UNKNOWN, vendor_len + 1);

	c->type = c->vendor + vendor_len + 1;
	memcpy((void *) c->type, type, type_len + 1);
	return 0;
}

/**
 * The parameters vendor and type are pointers to the compatible string,
 * both point to the beginning of the contents to be extracted.
 * It is expected that the whole compatible string looks like this:
 *   <vendor><SEP><type><NULL>
 * Allocates the extracted values at once, to free, just do free(c->vendor).
 */
static
int use_vendor_and_type(struct hwio_comp_spec *c, const char *vendor, const char *type)
{
	const size_t vendor_len = type - vendor - 1;

	c->vendor = strndup(vendor, strlen(vendor));
	if(c->vendor == NULL)
		return 1;

	c->vendor[vendor_len] = '\0';
	c->type = c->vendor + vendor_len + 1;
	return 0;
}

/**
 * The parameters type and version are pointers to the compatible string,
 * both point to the beginning of the contents to be extracted.
 * It is expected that the whole compatible looks like this:
 *   <type><SEP><version><NULL>
 *
 * The vendor is set to unknown.
 * Allocates the expected values at once, to free, just do free(c->vendor).
 */
static
int use_type_and_version(struct hwio_comp_spec *c, const char *type, const char *version)
{
	const size_t vendor_len = strlen(HWIO_VENDOR_UNKNOWN);
	const size_t type_len = version - type - 1;

	c->vendor = malloc(vendor_len + 1 + type_len + 1);
	if(c->vendor == NULL)
		return 2;

	memcpy((void *) c->vendor, HWIO_VENDOR_UNKNOWN, vendor_len + 1);

	c->type = c->vendor + vendor_len + 1;
	memcpy((void *) c->type, type, type_len);
	c->type[type_len] = '\0';

	parse_version(&c->version, version);
	return 0;
}

/**
 * The parameters vendor, type and version point to the compatible string.
 * It is expected that the whole compatible looks like this:
 *   <vendor><SEP><type><SEP><version>
 *
 * Allocates the expected values at once, to free, just do free(c->vendor).
 */
static
int use_vendor_type_version(struct hwio_comp_spec *c, const char *vendor, const char *type, const char *version)
{
	const size_t vendor_len = type - vendor - 1;
	const size_t type_len   = version - type - 1;

	c->vendor = strndup(vendor, vendor_len + 1 + type_len);
	if(c->vendor == NULL)
		return 1;

	c->vendor[vendor_len] = '\0';
	c->type = c->vendor + vendor_len + 1;

	parse_version(&c->version, version);
	return 0;
}

/**
 * Parses the given compat string into the spec instance.
 */
static
int parse_one_compat(struct hwio_comp_spec *spec, const char *compat)
{
	spec->vendor = NULL;
	spec->type   = NULL;
	hwio_version_set(&spec->version, HWIO_VERSION_NA, HWIO_VERSION_NA, HWIO_VERSION_NA);

	if(compat == NULL)
		return 0;

	const size_t len  = strlen(compat);

	// should look like this:
	//  <vendor>,<type>-<version>
	void *comma = memchr(compat, ',', len);
	void *hyphen = memrchr(compat, '-', len);

	if(comma == NULL && hyphen == NULL)
		return treat_it_as_type(spec, compat);

	else if(comma != NULL && hyphen == NULL)
		return use_vendor_and_type(spec, compat, ((char *) comma) + 1);

	else if(comma == NULL && hyphen != NULL)
		return use_type_and_version(spec, compat, ((char *) hyphen) + 1);

	else
		return use_vendor_type_version(spec, compat, ((char *) comma) + 1, ((char *) hyphen) + 1);
}

static
int parse_first_compat(struct hwio_comp *c, const char **compat)
{
	struct hwio_comp_spec spec;

	if(parse_one_compat(&spec, compat[0]))
		return 1;

	c->vendor = spec.vendor;
	c->type = spec.type;
	memcpy(&c->version, &spec.version, sizeof(struct hwio_version));

	return 0;
}

/**
 * Conversion of dtree_dev_t into hwio_comp instance.
 * The hwio_comp is allocated on the heap and fully initialized.
 */
static
struct hwio_comp *from_dtree_dev(const struct dtree_dev_t *d, const struct hwio_dev *dev)
{
	const char *dname    = dtree_dev_name(d);
	const uint32_t dbase = dtree_dev_base(d);
	const uint32_t dhigh = dtree_dev_high(d);

#define ALL_ONES ((uint32_t) ~0)
	// if the range is not provided, use the greatest range possible
	const uint32_t range = dhigh > dbase? dhigh - dbase : ALL_ONES - dbase;

	struct hwio_comp *c = hwio_comp_alloc(dname, dbase, range);
	if(c == NULL)
		return NULL;

	c->dev = dev;

	if(parse_first_compat(c, dtree_dev_compat(d))) {
		free(c); // take care when changing the hwio_comp_alloc
		return NULL;
	}

	c->plat_free = &hwio_linux_comp_free;
	return c;
}

static
int dtree_dev_matches(const struct dtree_dev_t *d,
		const struct hwio_comp_spec *qspec)
{
	const char **compat = dtree_dev_compat(d);

	for(int i = 0; compat[i] != NULL; ++i) {
		struct hwio_comp_spec spec = {NULL, NULL};

		if(parse_one_compat(&spec, compat[i]))
			continue; // failed, ignore, try next

		if(hwio_comp_spec_match(&spec, qspec)) {
			free(spec.vendor);
			return 1;
		}

		free(spec.vendor);
	}

	return 0;
}

static
int do_dtree_using_spec_and_close(struct hwio_dev *dev,
		const struct hwio_comp_spec *qspec, struct hwio_comp **head)
{
	dtree_reset();

	struct dtree_dev_t *d = NULL;
	struct hwio_comp *orig_head = *head;

	while((d = dtree_next())) {
		if(!dtree_dev_matches(d, qspec)) {
			dtree_dev_free(d);
			continue;
		}

		struct hwio_comp *comp = from_dtree_dev(d, dev);
		dtree_dev_free(d);

		if(comp == NULL)
			goto fail_memory;

		if (hwio_comp_append_if_unique(head, comp) != HWIO_EOK)
			goto fail_memory;
	}

	if(!dtree_iserror()) { // everything OK
		do_dtree_close(dev);
		return HWIO_EOK;
	}

fail_memory:
	hwio_comp_list_free(orig_head);
	do_dtree_close(dev);
	return HWIO_ENOMEM;
}

static
int do_dtree_full_query_and_close(struct hwio_dev *dev,
		const char *name, const struct hwio_comp_spec *qspec,
		struct hwio_comp **head)
{
	dtree_reset();

	struct dtree_dev_t *d = NULL;
	struct hwio_comp *orig_head = *head;

	while((d = dtree_byname(name))) {
		if(!dtree_dev_matches(d, qspec)) {
			dtree_dev_free(d);
			continue;
		}

		struct hwio_comp *comp = from_dtree_dev(d, dev);
		dtree_dev_free(d);

		if(comp == NULL)
			goto fail_memory;

		if (hwio_comp_append_if_unique(head, comp) != HWIO_EOK)
			goto fail_memory;
	}

	if(!dtree_iserror()) {
		do_dtree_close(dev);
		return HWIO_EOK;
	}

fail_memory:
	hwio_comp_list_free(orig_head);
	do_dtree_close(dev);
	return HWIO_ENOMEM;
}

int hwio_comp_find_spec(struct hwio_dev *dev,
	const struct hwio_comp_spec *spec, const char *name,
	struct hwio_comp **head)
{
	assert(spec != NULL);

	int ret = 0;
	if((ret = do_dtree_open(dev)))
		return ret;

	if(name == NULL)
		return do_dtree_using_spec_and_close(dev, spec, head);

	else
		return do_dtree_full_query_and_close(dev, name, spec, head);
}

///////////////////////////////

#ifdef TEST_HWIO_QUERY
#include "hwio_query.h"
#include "hwio_test.h"

void test_treat_it_as_type(void)
{
	test_start();
	struct hwio_comp_spec spec;
	int ret;

	if((ret = treat_it_as_type(&spec, "testing-type"))) {
		test_fail("Failed to use the compatible as a type only (malloc?)", ret);
		return;
	}

	test_assert(!strcmp(spec.vendor, HWIO_VENDOR_UNKNOWN), "Vendor doesn't match");
	test_assert(!strcmp(spec.type, "testing-type"), "Type doesn't match");

	free(spec.vendor);
	test_end();
}

void test_use_vendor_and_type(void)
{
	test_start();
	struct hwio_comp_spec spec;
	int ret;

	const char *compatible = "test-vendor,testing-type";
	const char *type = compatible + strlen("test-vendor") + 1;

	if((ret = use_vendor_and_type(&spec, compatible, type))) {
		test_fail("Failed to use the compatible as vendor and type (malloc?)", ret);
		return;
	}

	test_assert(!strcmp(spec.vendor, "test-vendor"), "Vendor doesn't match");
	test_assert(!strcmp(spec.type  , "testing-type"), "Type doesn't match");

	free(spec.vendor);
	test_end();
}

void test_use_type_and_version(void)
{
	test_start();
	struct hwio_comp_spec spec;
	int ret;

	const char *compatible = "testing-type-1.0.a";
	const char *version = compatible + strlen("testing-type") + 1;

	if((ret = use_type_and_version(&spec, compatible, version))) {
		test_fail("Failed to use the compatible as type and version (malloc?)", ret);
		return;
	}

	test_assert(!strcmp(spec.vendor, HWIO_VENDOR_UNKNOWN), "Vendor doesn't match");
	test_assert(!strcmp(spec.type, "testing-type"), "Type doesn't match");
	test_version(spec.version, 1, 0, 'a');

	free(spec.vendor);
	test_end();
}

void test_use_vendor_type_version(void)
{
	test_start();
	struct hwio_comp_spec spec;
	int ret;

	const char *compatible = "test-vendor,testing-type-1.0.a";
	const char *type = compatible + strlen("test-vendor") + 1;
	const char *version = type + strlen("testing-type") + 1;

	if((ret = use_vendor_type_version(&spec, compatible, type, version))) {
		test_fail("Failed to use the compatible as vendor, type and version (malloc?)", ret);
		return;
	}

	test_assert(!strcmp(spec.vendor, "test-vendor"), "Vendor doesn't match");
	test_assert(!strcmp(spec.type, "testing-type"), "Type doesn't match");
	test_version(spec.version, 1, 0, 'a');

	free(spec.vendor);
	test_end();
}

void test_parse_version(void)
{
	test_start();
	struct hwio_version ver;

	parse_version(&ver, "1.00.a");
	test_version(ver, 1, 0, 'a');

	parse_version(&ver, "2.00.c");
	test_version(ver, 2, 0, 'c');

	parse_version(&ver, "3.00.0");
	test_version(ver, 3, 0, 0);

	parse_version(&ver, "4.00");
	test_version(ver, 4, 0, HWIO_VERSION_NA);

	parse_version(&ver, "5.0");
	test_version(ver, 5, 0, HWIO_VERSION_NA);

	parse_version(&ver, "6.0.1");
	test_version(ver, 6, 0, 1);

	parse_version(&ver, "7.1.1");
	test_version(ver, 7, 1, 1);

	parse_version(&ver, "8.2.a");
	test_version(ver, 8, 2, 'a');

	parse_version(&ver, "12.12.c");
	test_version(ver, 12, 12, 'c');
	test_end();
}

void run_unit_tests(void)
{
	test_treat_it_as_type();
	test_use_vendor_and_type();
	test_use_type_and_version();
	test_use_vendor_type_version();
	test_parse_version();
}

int main(int argc, char **argv)
{
	run_unit_tests();

	const char *dtree_root = argc > 1? argv[1] : NULL;
	printf("dtree_root: %s\n", dtree_root);

	char *c_vendor = argc > 2? argv[2] : "xlnx";
	char *c_type   = argc > 3? argv[3] : "plb-v46";

	struct hwio_dev dev;
	int ret;

	subtest_start("hwio_init");
	if((ret = hwio_init(&dev, hwio_default_dev_phys(), dtree_root))) {
		perror("hwio_init");
		test_fail("Failed to hwio_init", ret);
		return 1;
	}
	subtest_end("hwio_init");

	struct hwio_comp_spec query[] = {
		{c_vendor, c_type, HWIO_VER3(0, 0, 0)},
		HWIO_COMPAT_END
	};
	struct hwio_comp *head = NULL;

	subtest_start("hwio_byspec");
	if((ret = hwio_byspec(&dev, query, &head))) {
		perror("hwio_byspec");
		test_fail("Failed to hwio_byspec", ret);
		return 2;
	}

	struct hwio_comp *curr;
	for(curr = head; curr != NULL; curr = hwio_comp_next(curr))
		printf("%s\n", hwio_comp_name(curr));
	subtest_end("hwio_byspec");

	hwio_comp_list_free(head);

	hwio_deinit(&dev);
	return 0;
}

#endif
