/**
 * @file hwio_linux_rw.c
 * @brief Implementation read/write from/to a component
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_linux.h"

#include "hwio_dev.h"
#include "hwio_comp.h"
#include "hwio_error.h"

#include <stdint.h>
#include <assert.h>

static
uint32_t *init_comp_access(struct hwio_comp *comp, uint32_t offset)
{
	assert(offset % 4 == 0);

	struct hwio_linux_dev *d  = as_linux_dev(comp->dev);
	struct hwio_linux_comp *c = as_linux_comp(comp);

	if(map_space_if_needed(d, c))
		return NULL;

	char *aligned_base = ((char *) c->mmap_m) + get_alignment(c->base);
	char *b = aligned_base + offset;
	return (uint32_t *) b;
}

int hwio_comp_write(struct hwio_comp *comp, uint32_t offset, uint32_t data)
{
	uint32_t *p = init_comp_access(comp, offset);
	if(p == NULL)
		return HWIO_EFAULT;

	*p = data;
	return 0;
}

int hwio_comp_read(struct hwio_comp *comp, uint32_t offset, uint32_t *data)
{
	uint32_t *p = init_comp_access(comp, offset);
	if(p == NULL)
		return HWIO_EFAULT;

	*data = *p;
	return 0;
}

#ifdef TEST_HWIO_RW
#include "hwio_test.h"

int main(int argc, char **argv)
{
	const char *file = argc > 1? argv[1] : "testing.dat";

	struct hwio_dev dev;
	int ret = 0;

	subtest_start("hwio_init");
	if((ret = hwio_init(&dev, file, ""))) {
		perror("hwio_init");
		test_fail("Call to hwio_init has failed", ret);
		return 1;
	}
	subtest_end("hwio_init");

	struct hwio_comp *c = hwio_comp_alloc("testing", 0x0000, 0x0016);
	c->dev = &dev;

	subtest_start("map_space_if_needed");
	if((ret = map_space_if_needed(as_linux_dev(&dev), as_linux_comp(c)))) {
		perror("map_space_if_needed");
		test_fail("Call to map_space_if_needed has failed", ret);
		goto failure;
	}
	subtest_end("map_space_if_needed");

	/////////////////////////////////////

	uint32_t data = 0xFFFFFFFF;

	subtest_start("hwio_comp_read from 0");
	if((ret = hwio_comp_read(c, 0, &data))) {
		perror("hwio_comp_read");
		test_fail("Call to hwio_comp_read has failed", ret);
		goto failure;
	}
	subtest_end("hwio_comp_read from 0");

	test_assert(data == 0, "Data in the file testing file are not zero");

	subtest_start("hwio_comp_write to 0");
	if((ret = hwio_comp_write(c, 0, 0xDEADBEEF))) {
		perror("hwio_comp_write");
		test_fail("Call to hwio_comp_write has failed", ret);
		goto failure;
	}
	subtest_end("hwio_comp_write to 0");

	subtest_start("hwio_comp_read from 0");
	if((ret = hwio_comp_read(c, 0, &data))) {
		perror("hwio_comp_read");
		test_fail("Call to hwio_comp_read has failed", ret);
		goto failure;
	}
	subtest_end("hwio_comp_read from 0");

	test_assert(data == 0xDEADBEEF, "Data in the file testing file are not zero");

failure:
	hwio_comp_free(c);
	hwio_deinit(&dev);
	return ret;
}
#endif
