/**
 * @file hwio_linux_util.c
 * @brief Implementation of utility functions
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_linux.h"
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/mman.h>
#include <fcntl.h>

struct hwio_comp *hwio_comp_alloc(const char *name,
		uint32_t base, uint32_t range)
{
	const size_t namesiz = strlen(name) + 1;
	const size_t msiz = sizeof(struct hwio_comp)
	                  + sizeof(struct hwio_linux_comp)
	                  + namesiz;

	// allocate all at once
	void *m = malloc(msiz);
	if(m == NULL)
		return NULL;

	struct hwio_comp *c = (struct hwio_comp *) m;
	struct hwio_linux_comp *plat = (struct hwio_linux_comp *) (c + 1);
	char *comp_name = (char *) (plat + 1);

	// init the c instance
	memset((void *) c, 0, sizeof(struct hwio_comp));

	// set plat_comp
	plat->base  = base;
	plat->range = range;
	plat->mmap_fd = -1;
	plat->mmap_m  = NULL;
	plat->mmap_siz = 0;
	c->plat_comp = plat;

	// set name
	c->name = (char *) memcpy((void *) comp_name, (void *) name, namesiz);

	return c;
}

void hwio_linux_comp_free(struct hwio_comp *c)
{
	if(c->vendor == NULL)
		return;

	free(c->vendor);
	c->vendor = NULL;
	c->type = NULL;

	unmap_space_if_mapped(as_linux_comp(c));
}

static
uint32_t get_page_count_to_mmap(uint32_t range)
{
	uint32_t count = range / getpagesize();
	return range % getpagesize() == 0? count : count + 1;
}

int map_space_if_needed(struct hwio_linux_dev *d, struct hwio_linux_comp *c)
{
	if(c->mmap_fd >= 0 && c->mmap_m != NULL)
		return 0;

	// both values are always consistent
	assert(c->mmap_fd < 0 && c->mmap_m == NULL);

	int fd = open(d->dev_file, O_RDWR);
	if(fd == -1)
		return -1;

	const uint32_t aligned_base = c->base - get_alignment(c->base);
	const uint32_t pagenum = get_page_count_to_mmap(c->range);
	const uint32_t mmapsiz = pagenum * getpagesize();
	const int prot_flags = PROT_READ | PROT_WRITE;

	void *m = mmap(NULL, mmapsiz, prot_flags, MAP_SHARED, fd, aligned_base);
	if(m == NULL) {
		close(fd);
		return -2;
	}

	c->mmap_fd  = fd;
	c->mmap_m   = m;
	c->mmap_siz = mmapsiz;

	return 0;
}

void unmap_space_if_mapped(struct hwio_linux_comp *c)
{
	if(c->mmap_fd < 0 && c->mmap_m == NULL)
		return;

	// both values are always consistent
	assert(c->mmap_fd >= 0 && c->mmap_m != NULL);

	munmap(c->mmap_m, c->mmap_siz);
	close(c->mmap_fd);

	c->mmap_m   = NULL;
	c->mmap_fd  = -1;
	c->mmap_siz = 0;
}
