/**
 * @file hwio.c
 * @brief Platform independent interface to HW
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#include "hwio_private.h"
#include "hwio.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#ifdef HWIO_DEBUG
#include <stdio.h>
#endif /* HWIO_DEBUG */

int hwio_version_match(struct hwio_version version,
	struct hwio_version pattern)
{
	if (pattern.major != HWIO_VERSION_NA &&
		pattern.major != version.major) {
		return 0;
	}

	if (pattern.minor != HWIO_VERSION_NA &&
		pattern.minor != version.minor ) {
		return 0;
	}

	if (pattern.subminor != HWIO_VERSION_NA &&
		pattern.subminor != version.subminor) {
		return 0;
	}

	return 1;
}

int hwio_comp_append_if_unique(struct hwio_comp **head,
		struct hwio_comp *item)
{
	struct hwio_comp_spec item_spec;
	if (hwio_comp_to_spec(item, &item_spec) != HWIO_EOK)
		return HWIO_EINVAL;

	struct hwio_comp *curr = *head;
	struct hwio_comp *last = NULL;

	while (curr != NULL) {
		struct hwio_comp_spec curr_spec;

		if (hwio_comp_to_spec(curr, &curr_spec) != HWIO_EOK)
			return HWIO_EINVAL;

		if (hwio_comp_name_and_spec_match(&curr_spec, curr->name,
			&item_spec, item->name))
			return HWIO_EOK;

		last = curr;
		curr = hwio_comp_next(curr);
	}

	hwio_comp_append(last, item);
	if (*head == NULL)
		*head = item;

	return HWIO_EOK;
}

void hwio_comp_free(struct hwio_comp *comp) 
{
	if(comp == NULL)
		return;

	if(comp->plat_free != NULL)
		comp->plat_free(comp);

	free(comp);
}

/**
 * Frees all the components in the linked list (comp->next)
 * but doesn't free the comp. The comp->next is set to NULL.
 */
static
void hwio_comp_free_next(struct hwio_comp *comp)
{
	struct hwio_comp *curr = hwio_comp_next(comp);

	while(curr != NULL) {
		struct hwio_comp *next = hwio_comp_next(curr);
		hwio_comp_free(curr);
		curr = next;
	}

	comp->next = NULL;
}

void hwio_comp_list_free(struct hwio_comp *head)
{
	if(head == NULL)
		return;

	hwio_comp_free_next(head);
	hwio_comp_free(head);
}

const char *hwio_comp_vendor(const struct hwio_comp *comp)
{
	return comp->vendor;
}

const char *hwio_comp_type(const struct hwio_comp *comp)
{
	return comp->type;
}

const char *hwio_comp_name(const struct hwio_comp *comp)
{
	return comp->name;
}

struct hwio_version hwio_comp_version(const struct hwio_comp *comp)
{
	return comp->version;
}

struct hwio_comp *hwio_comp_next(const struct hwio_comp *comp)
{
	return comp->next;
}

int hwio_comp_to_spec(const struct hwio_comp *comp, struct hwio_comp_spec *spec)
{
	if (comp == NULL || spec == NULL)
		return HWIO_EINVAL;

	spec->vendor = comp->vendor;
	spec->type = comp->type;
	spec->version = comp->version;

	return HWIO_EOK;
}

int hwio_comp_spec_match(const struct hwio_comp_spec *comp,
	const struct hwio_comp_spec *spec)
{
	if(comp == NULL)
		return 0;

	if(spec == NULL)
		return 1;

	if(spec->vendor != NULL && strcmp(comp->vendor, spec->vendor))
		return 0;

	if(spec->type != NULL && strcmp(comp->type, spec->type))
		return 0;

	return hwio_version_match(comp->version, spec->version);
}

int hwio_comp_name_and_spec_match(const struct hwio_comp_spec *comp,
	const char *comp_name, const struct hwio_comp_spec *spec,
	const char *spec_name)
{
	if (comp == NULL || comp_name == NULL)
		return 0;

	if(!hwio_comp_spec_match(comp, spec))
		return 0;

	if (spec_name != NULL && strcmp(comp_name, spec_name) != 0)
		return 0;

	return 1;
}

static
int hwio_compat_end(const struct hwio_comp_spec *spec)
{
	return (spec->vendor == NULL);
}

int hwio_byname(struct hwio_dev *dev, const char *name,
	const struct hwio_comp_spec *spec, struct hwio_comp **head)
{
	if (spec == NULL)
		return HWIO_EINVAL;

	unsigned int i;
	for (i = 0; !hwio_compat_end(&spec[i]); i++) {
		int retcode = hwio_comp_find_spec(dev, &spec[i], name, head);
		if (retcode != HWIO_EOK) {
			hwio_comp_list_free(*head);
			*head = NULL;
			return retcode;
		}
	}

	return HWIO_EOK;
}

int hwio_byspec(struct hwio_dev *dev, const struct hwio_comp_spec *spec,
	struct hwio_comp **head)
{
	return hwio_byname(dev, NULL, spec, head);
}

#ifdef TEST_HWIO_COMMON
#include "hwio_test.h"

void test_version_match(void)
{
	test_start();

	struct hwio_version v1;
	struct hwio_version v2;

	hwio_version_set(&v1, 4, 5, 6);
	hwio_version_set(&v2, 4, 5, 0);
	test_assert(!hwio_version_match(v1, v2), "Version matches");

	hwio_version_set(&v2, 4, 5, 6);
	test_assert(hwio_version_match(v1, v2), "Version does not match");

	hwio_version_set(&v2, 4, 5, HWIO_VERSION_NA);
	test_assert(hwio_version_match(v1, v2), "Version does not match");

	hwio_version_set(&v2, 4, HWIO_VERSION_NA, HWIO_VERSION_NA);
	test_assert(hwio_version_match(v1, v2), "Version does not match");

	hwio_version_set(&v2, HWIO_VERSION_NA, HWIO_VERSION_NA, HWIO_VERSION_NA);
	test_assert(hwio_version_match(v1, v2), "Version does not match");

	hwio_version_set(&v2, HWIO_VERSION_NA, 5, HWIO_VERSION_NA);

	test_end();
}

void test_comp_spec_match(void)
{
	test_start();

	struct hwio_comp_spec s1;
	struct hwio_comp_spec s2;

	s1 = (struct hwio_comp_spec)
		{ "test-vendor", "test-type-1", {1, 2, 0} };
	s2 = (struct hwio_comp_spec)
		{ "test-vendor", "test-type-2", {1, 2, 0} };
	test_assert(!hwio_comp_spec_match(&s1, &s2),
		"Specification matches");

	s2 = (struct hwio_comp_spec)
		{ "test-vendor", "test-type-1", {1, 2, 0} };
	test_assert(hwio_comp_spec_match(&s1, &s2),
		"Specification does not match");

	s2 = (struct hwio_comp_spec)
		{ "test-vendor", NULL, {1, 2, 0} };
	test_assert(hwio_comp_spec_match(&s1, &s2),
		"Specification does not match");

	s2 = (struct hwio_comp_spec)
		{ NULL, "test-type-1", {1, 2, 0} };
	test_assert(hwio_comp_spec_match(&s1, &s2),
		"Specification does not match");

	s2 = (struct hwio_comp_spec)
		{ NULL, NULL, {1, 2, 0} };
	test_assert(hwio_comp_spec_match(&s1, &s2),
		"Specification does not match");

	s2 = (struct hwio_comp_spec)
		{ "test-vendor", "test-type-1", {1, 2, HWIO_VERSION_NA} };
	test_assert(hwio_comp_spec_match(&s1, &s2),
		"Specification does not match");

	test_end();
}

void test_comp_name_and_spec_match()
{
	test_start();

	struct hwio_comp_spec s1;
	struct hwio_comp_spec s2;

	s1 = (struct hwio_comp_spec)
		{ "test-vendor", "test-type", {1, 2, 0} };
	s2 = (struct hwio_comp_spec)
		{ "test-vendor", "test-type", {1, 2, 0} };
	const char *name1 = "comp-name-1";
	const char *name2 = "comp-name-2";
	test_assert(!hwio_comp_name_and_spec_match(&s1, name1, &s2, name2),
		"Specification and name match");

	name2 = "comp-name-1";
	test_assert(hwio_comp_name_and_spec_match(&s1, name1, &s2, name2),
		"Specification and name do not match");

	name2 = NULL;
	test_assert(hwio_comp_name_and_spec_match(&s1, name1, &s2, name2),
		"Specification and name do not match");

	s2 = (struct hwio_comp_spec)
		{ "test-vendor", "test-type", {1, 2, 3} };
	test_assert(!hwio_comp_name_and_spec_match(&s1, name1, &s2, name2),
		"Specification and name match");

	test_end();
}

static
void hwio_spec_name_to_comp(const struct hwio_comp_spec *spec,
	char *name, struct hwio_comp *comp)
{
	memset(comp, 0, sizeof(*comp));

	comp->vendor = spec->vendor;
	comp->type = spec->type;
	comp->name = name;
	comp->version = spec->version;
}

/* Test implementation */
int hwio_comp_find_spec(struct hwio_dev *dev,
	const struct hwio_comp_spec *spec, const char *name,
	struct hwio_comp **head)
{
	static struct hwio_comp_spec comps[] = {
		{ "dsjfskdjf", "sdkfjsdlkfj", HWIO_VER2(9, 9) },
		{ "test-vendor-2", "test-comp-2", HWIO_VER3(1, 0 ,5) },
		{ "332904832", "34293483940", HWIO_VER3(1, 2, 3) },
		{ "test-vendor-1", "test-comp-1", HWIO_VER2(2, 3) },
		{ "j4jk3j5h5", "3j6n4b0d1e4", HWIO_VER3(4, 5, 6) },
		HWIO_COMPAT_END
	};

	struct hwio_comp_spec *curr;
	for (curr = comps; !hwio_compat_end(curr); curr++) {
		if (hwio_comp_spec_match(curr, spec)) {
			struct hwio_comp *comp = malloc(sizeof(*comp));
			hwio_spec_name_to_comp(curr, (char *) name, comp);
			hwio_comp_append_if_unique(head, comp);
		}
	}

	return HWIO_EOK;
}

void test_byname()
{
	test_start();

	struct hwio_comp_spec compat[] = {
		{ "test-vendor-1", "test-comp-1", HWIO_VER2(2, 3) },
		{ "test-vendor-2", "test-comp-2", HWIO_VER3(1, 0, HWIO_VERSION_NA)},
		HWIO_COMPAT_END
	};

	struct hwio_comp *head = NULL;
	const char *name = "test-name";
	int retcode = hwio_byname(NULL, name, compat, &head);
	test_assert(retcode == HWIO_EOK, "Error calling hwio_byname");
	test_assert(head != NULL, "No component found");

	struct hwio_comp_spec *curr_spec = compat;
	while (!hwio_compat_end(curr_spec)) {
		int found = 0;
		struct hwio_comp *curr_comp = head;

		while (curr_comp != NULL) {
			struct hwio_comp_spec comp_spec;
			hwio_comp_to_spec(curr_comp, &comp_spec);

			if (hwio_comp_spec_match(&comp_spec, curr_spec))
				found = 1;

			curr_comp = hwio_comp_next(curr_comp);
		}

		test_assert(found, "Component not found");
		curr_spec++;
	}

	hwio_comp_list_free(head);

	test_end();
}

void run_unit_tests(void)
{
	test_version_match();
	test_comp_spec_match();
	test_comp_name_and_spec_match();
	test_byname();
}

int main(int argc, char *argv[])
{
	run_unit_tests();
	return 0;
}

#endif /* TEST_HWIO_COMMON */
