/**
 * @file hwio_comp.h
 * @brief Interface of a component (Public API)
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#ifndef _HWIO_COMP_H
#define _HWIO_COMP_H

#include "hwio_version.h"
#include "hwio_dev.h"
#include "hwio_comp_spec.h"

#include <stddef.h>
#include <stdint.h>

#define HWIO_ADDR_SPACE_NA 0

/**
 * Component structure. Forward declaration.
 */
struct hwio_comp;

/**
 * Pointer type to component deallocation function.
 */
typedef void (*hwio_plat_comp_free_t)(struct hwio_comp *);

/**
 * Free the component. Call platform-dependent free-function
 * of the component and then free the component stucture.
 */
void hwio_comp_free(struct hwio_comp *comp);

/**
 * Free all the components in the linked list.
 */
void hwio_comp_list_free(struct hwio_comp *head);

/**
 * Get the vendor of the component. Access function.
 * The vendor is never NULL. For an unknown vendor the
 * HWIO_VENDOR_UNKNOWN value is used.
 */
const char *hwio_comp_vendor(const struct hwio_comp *comp);

/**
 * Get the type of the component. Access function.
 * The type is never NULL. For an unknown type the HWIO_TYPE_UNKNOWN
 * value is used.
 */
const char *hwio_comp_type(const struct hwio_comp *comp);

/**
 * Get the name of the component. Access function.
 */
const char *hwio_comp_name(const struct hwio_comp *comp);

/**
 * Get the version of the component. Access function.
 */
struct hwio_version hwio_comp_version(const struct hwio_comp *comp);

/**
 * Get the successor of the component. Access function.
 */
struct hwio_comp *hwio_comp_next(const struct hwio_comp *comp);

/**
 * Write data into the component. Platform dependent.
 * @param offset The offset relative to the component's base. It must be
 *   a multiple of 4. If the resulting address (component's base + offset)
 *   is not a multiple of 4, the behaviour is undefined.
 * @return HWIO_EFAULT when trying to write out of the address space
 *   of the component (if this information is available), HWIO_EINVAL
 *   when the offset is not a multiple of 4, HWIO_EOK otherwise.
 */
int hwio_comp_write(struct hwio_comp *comp, uint32_t offset, uint32_t data);

/**
 * Read data from the component. Platform dependent.
 * @param offset The offset relative to the component's base. It must be
 *   a multiple of 4. If the resulting address (component's base + offset)
 *   is not a multiple of 4, the behaviour is undefined.
 * @return HWIO_EFAULT when trying to read out of the address space
 *   of the component (if this information is available), HWIO_EINVAL
 *   when the offset is not a multiple of 4, HWIO_EOK otherwise.
 */
int hwio_comp_read(struct hwio_comp *comp, uint32_t offset, uint32_t *data);

#endif /* _HWIO_COMP_H */
