/**
 * @file hwio_private.h
 * @brief Private API used inside the library only
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @author Denis Matoušek <xmatou08@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 */

#ifndef _HWIO_PRIVATE_H
#define _HWIO_PRIVATE_H

#include "hwio_dev.h"
#include "hwio_comp.h"
#include "hwio_comp_spec.h"

#include <stdlib.h>
#include <assert.h>

/**
 * Component structure.
 */
struct hwio_comp {
	void *plat_comp; ///< Platform-dependent component data
	hwio_plat_comp_free_t plat_free; ///< Platform-dependent deallocation
	const struct hwio_dev *dev; ///< Platform-independent device data
	char *vendor;
	char *type;
	char *name;
	struct hwio_version version;
	struct hwio_comp *next;
};

/**
 * Appends the given item to the given list (tail).
 * Returns the new tail.
 */
static inline
struct hwio_comp *hwio_comp_append(struct hwio_comp *tail,
		struct hwio_comp *item)
{
	assert(tail == NULL || tail->next == NULL);

	item->next = NULL;

	if(tail != NULL)
		tail->next = item;

	return item;
}

/**
 * Appends the given item to the given list only if the item
 * is not already contained in the list.
 * @param head If *head is NULL, new list (its first item) is returned via
 *   this parameter. Otherwise, the pointer is left unchanged and the item
 *   is appended to the end of the list (if it is unique).
 * @item item New item.
 * @return HWIO_EINVAL on error, HWIO_EOK on success.
 */
int hwio_comp_append_if_unique(struct hwio_comp **head,
		struct hwio_comp *item);

/**
 * Copy data from `struct hwio_comp' to `struct hwio_comp_spec'.
 */
int hwio_comp_to_spec(const struct hwio_comp *comp, struct hwio_comp_spec *spec);

/**
 * Find components in the device that match the specification and the name.
 * @param spec Component specification.
 * @param name Component name.
 * @param head The first item of the resulting list. Intended for concatenation.
 * @return HWIO_EOK on success, HWIO_ENOMEM on allocation error
 *   or HWIO_EINVAL on invalid argument.
 */
int hwio_comp_find_spec(struct hwio_dev *dev,
	const struct hwio_comp_spec *spec, const char *name,
	struct hwio_comp **head);

/**
 * Checks if the component specification matchs the given
 * specification.
 * @param comp Component specification.
 * @param spec Specification to match. If some of its fields are NULL, that one matches.
 */
int hwio_comp_spec_match(const struct hwio_comp_spec *comp,
	const struct hwio_comp_spec *spec);

/**
 * Check if the component specification and its name match
 * other specification and name.
 * @param comp Component specification.
 * @param comp_name Component name.
 * @param spec Specification. If some of its fields is NULL, this field matches.
 * @param spec_name Specification name. If NULL, it matches.
 */
int hwio_comp_name_and_spec_match(const struct hwio_comp_spec *comp,
	const char *comp_name, const struct hwio_comp_spec *spec,
	const char *spec_name);

#endif
