/**
 * @file hwio_tool.h
 * @brief Helpers for tools made on top of hwio. (Public)
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * @mainpage
 *
 * The library libhwio is a part of the sec6net project
 * https://merlin.fit.vutbr.cz/wiki-sec6net/.
 */

#ifndef __HWIO_TOOL_H__
#define __HWIO_TOOL_H__

#include "hwio_comp_spec.h"
#include "hwio_comp.h"

#include <stdarg.h>
#include <string.h>
#include <stdint.h>

//////////////////////////
// Command line parsing //
//////////////////////////

/**
 * General options accapted by all tools.
 * Not all of them are mandatory to be implemented, but
 * all of them can be parsed.
 */
#define HWIO_TOOL_OPTS "hvqx:d:ALIE:R"

/**
 * Structure to hold the common options.
 */
struct hwio_tool_opts {
	int verbose;
	int all_comps;
	const char *dev_src;
	const char *dev_phys;
};

/**
 * Structure to hold information about a command.
 */
struct hwio_tool_cmd {
	int code;
	const char *arg;
};

/**
 * Forward declaration of a structure used by the
 * tool. The tool can customize its contents.
 */
struct hwio_tool_custom;

/**
 * Use this macro to simply avoid using of hwio_tool_custom.
 */
#define HWIO_TOOL_CUSTOM_NOT_USED \
	struct hwio_tool_custom {void *empty;};

/**
 * When hwio_tool_custom is not used, then use this as a NULL reference.
 * Applies especially for hwio_tool_params_new.
 */
#define HWIO_TOOL_CUSTOM_NULL ((struct hwio_tool_custom *) 0)

/**
 * Cover of the command-line related structures.
 *
 * It stores index to argv of the first component name (first_name).
 * It also stores the argv and argc (usually from the main function).
 */
struct hwio_tool_params {
	int first_name;
	char * const *argv;
	int argc;
	struct hwio_tool_opts opts;
	struct hwio_tool_cmd  cmd;
	struct hwio_tool_custom *custom;
};

/**
 * Initialization of the hwio_tool_params.
 */
static inline
struct hwio_tool_params hwio_tool_params_new(struct hwio_tool_custom *c,
		int argc, char * const argv[])
{
	struct hwio_tool_params params;
	memset(&params, 0, sizeof(struct hwio_tool_params));
	params.custom = c;
	params.argv = argv;
	params.argc = argc;
	return params;
}

/**
 * Tests whether the cmd is already set.
 */
static inline
int hwio_tool_params_has_cmd(const struct hwio_tool_params *p)
{
	return p->cmd.code != 0;
}

/**
 * Function to be implemented by the `tool`.
 * It parses the given option and modifies the params appropriately.
 * It should access mostly the custom part.
 *
 * The first arg (opt) holds the recognized option. If the option accepts
 * a parameter, the second arg (optarg) holds its value. Otherwise it is
 * NULL (the same if the optarg is optional and not provided).
 *
 * The return value must be generated by the hwio_tool_custom_* helping
 * functions/macros.
 */
typedef int (*hwio_tool_custom_parse_t)(int /*opt*/, const char * /*optarg*/,
		struct hwio_tool_params * /*params*/);

/**
 * Custom option parsing was successful.
 */
#define HWIO_TOOL_CUSTOM_OK       0

/**
 * Custom option is not known to the parser.
 */
#define HWIO_TOOL_CUSTOM_UNKNOWN -1

/**
 * Just exit the tool after command line parsing.
 * No other processing will happen, but no error
 * has occured.
 */
#define HWIO_TOOL_CUSTOM_EXIT     1

/**
 * The function wraps the getopt processing. It fills the params
 * appropriately. The custom_parser is used when an option is not
 * recognized (when it is not a general option).
 * The optstring contains the options specification of the `tool`
 * without the general ones.
 *
 * The function prints appropriate error messages if needed.
 *
 * It returns HWIO_TOOL_PARAMS_* values:
 * * OK     when the command line is OK
 * * FAILED when an option is invalid or there is missing an optarg
 * * EXIT   when an option forces the tool to exit successfully (help)
 */
int hwio_tool_params_parse(struct hwio_tool_params *params,
		hwio_tool_custom_parse_t custom_parser, const char *optstring);

#define HWIO_TOOL_PARAMS_OK     0
#define HWIO_TOOL_PARAMS_FAILED -1
#define HWIO_TOOL_PARAMS_EXIT   1

/**
 * Prints the help for this tool.
 * This is to be implemented in the tool.
 */
void hwio_tool_help(const char *toolname,
		const struct hwio_tool_params *params);


////////////////////////////
// Logging/Error handling //
////////////////////////////

/**
 * Error during the custom command line parsing.
 * Returns error code that says that something has failed.
 *
 * The message may be modified with a prefix (like a date
 * or something) and a newline is always appended.
 */
int hwio_tool_custom_failed(int opt, const char *msg, ...);

/**
 * Prints error and returns a suitable error code for the situation
 * when a command is already defined.
 */
#define hwio_tool_error_another_cmd(opt, params) \
	hwio_tool_custom_failed((opt), "command -%c was given already", (params)->cmd.code);

/**
 * Function that prints an error message and returns the given
 * error code.
 *
 * The message may be modified with a prefix (like a date
 * or something) and a newline is always appended.
 */
int hwio_tool_error(int errcode, const char *fmt, ...);

/**
 * Verbose implementation.
 * The message is printed when verbose > params->opts.verbose.
 * The verbose must always be greater or equal to 0.
 *
 * The message may be modified with a prefix (like a date
 * or something) and a newline is always appended.
 */
void _hwio_tool_verbose(const struct hwio_tool_params *params,
		int verbose, const char *func, const char *file, int line,
		const char *msg, ...);

/**
 * Wrapper around _hwio_tool_verbose.
 * The third argument MUST be the string format.
 */
#define hwio_tool_verbose(p, v, ...) \
	_hwio_tool_verbose((p), (v), __func__, __FILE__, __LINE__, __VA_ARGS__)

/**
 * Prints the given message and aborts execution.
 * It should not be called directly.
 */
void _hwio_tool_die(const char *func, const char *file, int line,
		const char *fmt, ...);

/**
 * Wrapper around _hwio_tool_die.
 * The first argument MUST be the string format.
 */
#define hwio_tool_die(...) \
	_hwio_tool_die(__func__, __FILE__, __LINE__, __VA_ARGS__)

/**
 * Prints "<label>: <value>\n".
 */
void hwio_tool_print_string(const char *label, const char *value);

/**
 * Prints "<label>: <value>\n".
 * The flags can modify the output. When the flags is zero than
 * the value is treated as 32 bit integer to be printed in decimal.
 */
void hwio_tool_print_uint(const char *label, uint32_t value, int flags);

/**
 * When bit[0] is set and bit[1] is unset, then the
 * value is treated to be 8 bit wide.
 */
#define HWIO_TOOL_UINT_8   0x0001

/**
 * When bit[1] is set and bit[0] is unset, then the
 * value is treated to be 16 bit wide.
 */
#define HWIO_TOOL_UINT_16  0x0002

/**
 * When the bit[2] is set, then the value is printed
 * as hexadecimal.
 */
#define HWIO_TOOL_UINT_HEX 0x0004


/////////////////////////////////////
// High-level wrappers around hwio //
/////////////////////////////////////

/**
 * Error handler.
 */
typedef void (*hwio_tool_comp_err_t)(struct hwio_comp *c, const char *fmt, ...);

/**
 * Default error handler suitable for most command line tools.
 */
void hwio_tool_error_handler(struct hwio_comp *c, const char *fmt, ...);

/**
 * An operation that can be performed on a component.
 * Returns one of HWIO_TOOL_OP_* constants.
 */
typedef int (*hwio_tool_comp_op_t)(struct hwio_comp *c, hwio_tool_comp_err_t error_handler,
		const struct hwio_tool_params *params);

#define HWIO_TOOL_OP_OK      0
#define HWIO_TOOL_OP_FAILED -1
#define HWIO_TOOL_OP_EXIT    1

/**
 * Executes an operation on the set of components.
 *
 * The operation is expected to return one of the HWIO_TOOL_OP_* constants.
 *
 * When HWIO_TOOL_OP_EXIT is returned then no more components are used
 * and the hwio_tool_exec_generic exits immediately.
 * When HWIO_TOOL_OP_FAILED is returned then the other components are
 * evaluated. But the return value will show that at least one error.
 *
 * Errors are reported using the given error_handler. If NULL is given
 * instead of an error_handler, then no error is reported at all.
 *
 * Returns 0 when all operations on all components are successful.
 * Otherwise it returns a standard error code.
 */
int hwio_tool_exec_generic(const struct hwio_tool_params *params,
		const struct hwio_comp_spec *spec, hwio_tool_comp_op_t op,
		hwio_tool_comp_err_t error_handler);

/**
 * Prints the basic info about a component (name, type, version, vendor, etc.).
 * in the format suitable for -I command.
 * It can be used as hwio_tool_comp_op_t. It always returns HWIO_TOOL_OP_OK.
 */
int hwio_tool_comp_info(struct hwio_comp *c, hwio_tool_comp_err_t err, const struct hwio_tool_params *params);

/**
 * Prints the name of the component. Useful to implement the -L command.
 * It always returns HWIO_TOOL_OP_OK.
 */
int hwio_tool_comp_name(struct hwio_comp *c, hwio_tool_comp_err_t err, const struct hwio_tool_params *params);


//////////////////
// Return codes //
//////////////////

/**
 * The tool was successful.
 */
#define HWIO_TOOL_EOK   0

/**
 * Something has failed.
 */
#define HWIO_TOOL_EFAIL 1

/**
 * The command line arguments are invalid.
 */
#define HWIO_TOOL_EARGS 2

#endif
