/**
 * @file hwio_tool_logging.c
 * @brief Implementation of logging/error printing helpers.
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * @mainpage
 *
 * The library libhwio is a part of the sec6net project
 * https://merlin.fit.vutbr.cz/wiki-sec6net/.
 */

#include "hwio_tool.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

static
void error_vprintf(const char *fmt, va_list args)
{
	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");
}

static
void location_print(const char *func, const char *file, int line)
{
	fprintf(stderr, "%s (%s:%d): ", func, file, line);
}

int hwio_tool_custom_failed(int opt, const char *msg, ...)
{
	fprintf(stderr, "While parsing -%c: ", opt);

	va_list args;
	va_start(args, msg);
	error_vprintf(msg, args);
	va_end(args);

	return -2;
}

int hwio_tool_error(int errcode, const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	error_vprintf(fmt, args);
	va_end(args);

	return errcode;
}

void _hwio_tool_verbose(const struct hwio_tool_params *params,
		int verbose, const char *func, const char *file, int line,
		const char *msg, ...)
{
	if(verbose > params->opts.verbose)
		return;

	location_print(func, file, line);
	va_list args;
	va_start(args, msg);
	error_vprintf(msg, args);
	va_end(args);
}

void _hwio_tool_die(const char *func, const char *file, int line,
		const char *fmt, ...)
{
	location_print(func, file, line);
	va_list args;
	va_start(args, fmt);
	error_vprintf(fmt, args);
	va_end(args);

	abort();
}

static
const char *skip_stars(const char *label)
{
	int i;

	for(i = 0; label[i] != '\0'; ++i) {
		if(label[i] == '*')
			continue;
	}

	return label + i;
}

void hwio_tool_print_string(const char *label, const char *value)
{
	printf("%s: %s\n", skip_stars(label), value);
}

void hwio_tool_print_uint(const char *label, uint32_t value, int flags)
{
	printf("%s: ", skip_stars(label));
	const int is_hex = flags & HWIO_TOOL_UINT_HEX;

#define HEX8  "0x%.02" PRIx8
#define HEX16 "0x%.04" PRIx16
#define HEX32 "0x%.08" PRIx32
#define DEC8  "%" PRIu8
#define DEC16 "%" PRIu16
#define DEC32 "%" PRIu32

	switch(flags & 0x3) {
	case HWIO_TOOL_UINT_8:
		printf(is_hex? HEX8 : DEC8, (uint8_t) value);
		break;

	case HWIO_TOOL_UINT_16:
		printf(is_hex? HEX16 : DEC16, (uint16_t) value);
		break;

	default:
		printf(is_hex? HEX32 : DEC32, (uint32_t) value);
		break;
	}

	printf("\n");
}
