/**
 * @file hwio_tool_params.c
 * @brief Implementation of command line parsing for tools made on top of hwio.
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * @mainpage
 *
 * The library libhwio is a part of the sec6net project
 * https://merlin.fit.vutbr.cz/wiki-sec6net/.
 */

#define _XOPEN_SOURCE

#include "hwio_tool.h"

#include <unistd.h>
#include <string.h>

/**
 * The function chooses between creating the optstring on fly or to
 * use the given one. If the given one contains the HWIO_TOOL_OPTS
 * then just use it. If it doesn't, then create a new one that contains
 * the given optstring and HWIO_TOOL_OPTS.
 */
static
const char *determine_opts_fmt(char *optsbuff, const char *optstring)
{
	if(strstr(optstring, HWIO_TOOL_OPTS) == NULL) {
		const size_t optstring_len = strlen(optstring);
		memcpy(optsbuff, optstring, optstring_len);
		memcpy(optsbuff + optstring_len, HWIO_TOOL_OPTS, strlen(HWIO_TOOL_OPTS) + 1);
		return optsbuff;
	}
	else {
		return optstring;
	}
}

int hwio_tool_params_parse(struct hwio_tool_params *params,
		hwio_tool_custom_parse_t custom_parser, const char *optstring)
{
	const int argc = params->argc;
	char * const *argv = params->argv;

	const size_t optstring_len = strlen(optstring);
	char optsbuff[optstring_len + strlen(HWIO_TOOL_OPTS) + 1];
	const char *opts_fmt = determine_opts_fmt(optsbuff, optstring);

	int opt;
	optind = 1;
	while((opt = getopt(argc, argv, opts_fmt)) != -1) {
		switch(opt) {
		case 'h':
			hwio_tool_help(argv[0], params);
			return HWIO_TOOL_PARAMS_EXIT;

		case 'v':
			params->opts.verbose += 1;
			break;

		case 'q':
			params->opts.verbose = -1;
			break;

		case 'x':
			params->opts.dev_src = optarg;
			break;

		case 'd':
			params->opts.dev_phys = optarg;
			break;

		case 'A':
			if(params->cmd.code == 'L') {
				hwio_tool_custom_failed(opt, "It doesn't make sense to combine -L and -A options");
				return HWIO_TOOL_PARAMS_FAILED;
			}

			params->opts.all_comps = 1;
			break;

		case 'L':
		case 'I':
		case 'R':
			if(hwio_tool_params_has_cmd(params)) {
				hwio_tool_error_another_cmd(opt, params)
				return HWIO_TOOL_PARAMS_FAILED;
			}

			params->cmd.code = opt;

			if(opt == 'L') {
				if(params->opts.all_comps) {
					hwio_tool_custom_failed(opt, "It doesn't make sense to combine -L and -A options");
					return HWIO_TOOL_PARAMS_FAILED;
				}

				params->opts.all_comps = 1; // listing always all components
			}

			break;

		case 'E':
			if(hwio_tool_params_has_cmd(params)) {
				hwio_tool_error_another_cmd(opt, params);
				return HWIO_TOOL_PARAMS_FAILED;
			}

			params->cmd.code = opt;
			params->cmd.arg = optarg;
			break;

		case '?':
			return HWIO_TOOL_PARAMS_FAILED; // getopt prints the message

		default:
			if(custom_parser == NULL)
				hwio_tool_die("BUG: no custom parser given, but the option -%c is unknown", opt);

			int err = custom_parser(opt, optarg, params);
			if(err == HWIO_TOOL_CUSTOM_UNKNOWN)
				hwio_tool_die("BUG: custom parser is not able to parse the option -%c", opt);

			else if(err > 0)
				return HWIO_TOOL_PARAMS_EXIT;

			else if(err == 0)
				continue;

			else
				return HWIO_TOOL_PARAMS_FAILED;
		}
	}

	if(params->opts.dev_phys == NULL)
		params->opts.dev_phys = hwio_default_dev_phys();

	hwio_tool_verbose(params, 1, "Device phys: %s", params->opts.dev_phys);

	if(params->opts.dev_src == NULL)
		params->opts.dev_src = hwio_default_dev_src();

	hwio_tool_verbose(params, 1, "Device src: %s", params->opts.dev_src);

	if(optind < argc && params->cmd.code == 'L') {
		hwio_tool_custom_failed('L', "It doesn't make sense to specify components for the command -L");
		return HWIO_TOOL_PARAMS_FAILED;
	}

	params->first_name = optind >= argc? argc : optind;
	return HWIO_TOOL_PARAMS_OK;
}
