/**
 * @file hwio_tool_wrappers.h
 * @brief Implementation of helpers for tools made on top of hwio.
 * @author Jan Viktorin <xvikto03@stud.fit.vutbr.cz>
 * @copyright Brno University of Technology
 * @date 2012
 *
 * @section license LICENSE TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this
 * product may be distributed under the terms of the GNU General Public
 * License (GPL), in which case the provisions of the GPL apply INSTEAD
 * OF those given above.
 *
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even
 * if advised of the possibility of such damage.
 *
 * @mainpage
 *
 * The library libhwio is a part of the sec6net project
 * https://merlin.fit.vutbr.cz/wiki-sec6net/.
 */

#include "hwio_tool.h"
#include "hwio.h"

#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

void hwio_tool_error_handler(struct hwio_comp *c, const char *fmt, ...)
{
	fprintf(stderr, "Error at %s: ", hwio_comp_name(c));

	va_list args;
	va_start(args, fmt);
	fprintf(stderr, fmt, args);
	va_end(args);

	fprintf(stderr, "\n");
}

static
int process_errcode(const char *context, int err)
{
	if(err == 0)
		return HWIO_TOOL_EOK;

	switch(err) {
	case HWIO_ENOMEM:
		return hwio_tool_error(HWIO_TOOL_EFAIL, "%s: Out of memory", context);

	case HWIO_EFAULT:
		return hwio_tool_error(HWIO_TOOL_EFAIL, "%s: Bad address", context);

	case HWIO_EINVAL:
		return hwio_tool_error(HWIO_TOOL_EFAIL, "%s: Invalid argument", context);

	case HWIO_EINFO:
		return hwio_tool_error(HWIO_TOOL_EFAIL, "%s: Failed to obtain device info", context);

	case HWIO_EATT:
		return hwio_tool_error(HWIO_TOOL_EFAIL, "%s: Failed to attach device", context);
	}

	return hwio_tool_error(HWIO_TOOL_EFAIL, "%s: Unknown error (a buggy error code?)", context);
}

int dev_init(struct hwio_dev *dev, const struct hwio_tool_params *params)
{
	assert(params->opts.dev_phys != NULL);
	assert(params->opts.dev_src != NULL);

	int err = hwio_init(dev, params->opts.dev_phys, params->opts.dev_src);
	return process_errcode("device init", err);
}

static
int many_bynames(struct hwio_dev *dev, const struct hwio_comp_spec *spec,
		struct hwio_comp **head, const struct hwio_tool_params *params)
{
	char * const *names = params->argv + params->first_name;
	const int count = params->argc - params->first_name;

	for(int i = 0; i < count; ++i) {
		int err;

		if((err = hwio_byname(dev, names[i], spec, head))) {
			hwio_comp_list_free(*head);
			return process_errcode("hwio_byname", err);
		}
	}

	return 0;
}

static
int comp_lookup(struct hwio_dev *dev, const struct hwio_comp_spec *spec,
		struct hwio_comp **head, const struct hwio_tool_params *params)
{
	if(params->opts.all_comps)
		return process_errcode("hwio_byspec", hwio_byspec(dev, spec, head));

	else if(params->first_name < params->argc)
		return many_bynames(dev, spec, head, params);

	else /* expect that just one component exist */ {
		int err = hwio_byspec(dev, spec, head);
		if(err)
			return process_errcode("hwio_byspec", err);

		if(head == NULL || *head == NULL)
			return 0; // no unit found, that's ok

		if(hwio_comp_next(*head) != NULL) {
			hwio_comp_list_free(*head);
			*head = NULL;
			return process_errcode("lookup one", HWIO_EINVAL); // more units, that is not what we want
		}
		else
			return 0; // got exactly 1 unit
	}
}

static
int comp_walk(struct hwio_comp *head, hwio_tool_comp_op_t op,
		hwio_tool_comp_err_t error_handler, const struct hwio_tool_params *params)
{
	if(head == NULL)
		return 0;

	int failures = 0;

	struct hwio_comp *curr = head;
	for(; curr != NULL; curr = hwio_comp_next(curr)) {
		int ret = op(curr, error_handler, params);

		switch(ret) {
		case HWIO_TOOL_OP_EXIT:
			goto break_for;

		case HWIO_TOOL_OP_FAILED:
			failures += 1;
			break;

		default:
			continue;
		}
	}

break_for:
	return failures;
}

int hwio_tool_exec_generic(const struct hwio_tool_params *params,
		const struct hwio_comp_spec *spec, hwio_tool_comp_op_t op,
		hwio_tool_comp_err_t error_handler)
{
	struct hwio_dev dev;
	int err;

	if((err = dev_init(&dev, params)))
		return err;

	struct hwio_comp *head = NULL;
	if((err = comp_lookup(&dev, spec, &head, params))) {
		hwio_deinit(&dev);
		return err;
	}

	err = comp_walk(head, op, error_handler, params);

	hwio_comp_list_free(head);
	hwio_deinit(&dev);

	return (err == 0)? HWIO_EOK :
		hwio_tool_error(HWIO_TOOL_EFAIL, "An operation failed to be performed on a component");
}

int hwio_tool_comp_info(struct hwio_comp *c,
		hwio_tool_comp_err_t err, const struct hwio_tool_params *params)
{
	printf("* %s", hwio_comp_name(c));
	printf(" %s:%s", hwio_comp_vendor(c), hwio_comp_type(c));

	struct hwio_version ver = hwio_comp_version(c);
	printf(":%d.%d", ver.major, ver.minor);

	if(ver.subminor == HWIO_VERSION_NA)
	printf("\n");
	else if(isalpha(ver.subminor) && islower(ver.subminor))
	printf(".%c\n", ver.subminor);
	else if(isdigit(ver.subminor))
	printf(".%c\n", ver.subminor);
	else
	printf(".%d\n", ver.subminor);

	return HWIO_TOOL_OP_OK;
}

int hwio_tool_comp_name(struct hwio_comp *c,
		hwio_tool_comp_err_t err, const struct hwio_tool_params *params)
{
	printf("%s\n", hwio_comp_name(c));
	return HWIO_TOOL_OP_OK;
}
